

/*  LibStart.c  $ 28/03/98 MT $                         */
/*                                                      */
/*  Library startup code and function table definition  */
/*  Adapted from a source code by Andreas R. Kleinert   */


#include "exec/types.h"
#include "exec/memory.h"
#include "exec/libraries.h"
#include "exec/execbase.h"
#include "exec/resident.h"
#include "exec/initializers.h"
#include "clib/exec_protos.h"
//include "pragmas/exec_sysbase_pragmas.h"
#include "pragmas/exec_pragmas.h"
#include "librarybase.h"
#include "libraryprotos.h"


#ifndef LIBRARYNAME
#define LIBRARYNAME "generic.library"
#endif

#ifndef LIBRARYID
#define LIBRARYID "generic 40.11 (28.3.98)"
#endif

#ifndef LIBRARYVERSION
#define LIBRARYVERSION 40
#endif

#ifndef LIBRARYREVISION
#define LIBRARYREVISION 11
#endif


struct MyDataInit
{
   UWORD ln_Type_Init;      UWORD ln_Type_Offset;      UWORD ln_Type_Content;
   UBYTE ln_Name_Init;      UBYTE ln_Name_Offset;      ULONG ln_Name_Content;
   UWORD lib_Flags_Init;    UWORD lib_Flags_Offset;    UWORD lib_Flags_Content;
   UWORD lib_Version_Init;  UWORD lib_Version_Offset;  UWORD lib_Version_Content;
   UWORD lib_Revision_Init; UWORD lib_Revision_Offset; UWORD lib_Revision_Content;
   UBYTE lib_IdString_Init; UBYTE lib_IdString_Offset; ULONG lib_IdString_Content;
   ULONG ENDMARK;
};

struct InitTable
{
   ULONG              LibBaseSize;
   APTR              *FunctionTable;
   struct MyDataInit *DataTable;
   APTR               InitLibTable;
};


extern ULONG __saveds __asm __UserLibInit(register __a0 struct LibraryBase *);
extern ULONG __saveds __asm __UserLibCleanup(register __a0 struct LibraryBase *);

struct LibraryBase * __saveds __asm InitLib(register __a6 struct ExecBase *, register __a0 SEGLISTPTR, register __d0 struct LibraryBase *);

struct LibraryBase * __saveds __asm OpenLib(register __a6 struct LibraryBase *);
SEGLISTPTR __saveds __asm CloseLib(register __a6 struct LibraryBase *);
SEGLISTPTR __saveds __asm ExpungeLib(register __a6 struct LibraryBase *);
ULONG __asm ExtFuncLib(void);

const char MyLibVersTag[] = LIBVERSTAG;
char __aligned MyLibName[] = LIBRARYNAME;
char __aligned MyLibID[] = LIBRARYID;


// ****************************************************************************************
// **** LibStart():                                                                    ****
// ****                                                                                ****
// **** If someone tries to start a library as an executable, it must return (LONG) -1 ****
// **** as result. That's what we are doing here.                                      ****
// ****************************************************************************************

LONG __asm LibStart(void)
{
   return (-1);
}


// ***************************************************************************************************
// **** ROMTag and Library initialization structure:                                              ****
// ****                                                                                           ****
// **** Below you find the ROMTag, which is the most important "magic" part of a library          ****
// **** (as for any other resident module). You should not need to modify any of the              ****
// **** structures directly, since all the data is referenced from constants from somewhere else. ****
// ****                                                                                           ****
// **** You may place the ROMTag directly after the LibStart (-> LibStart.c) function as well.    ****
// ****                                                                                           ****
// **** Note that the data initialization structure may be somewhat redundant - it's              ****
// **** for demonstration purposes.                                                               ****
// ****                                                                                           ****
// **** EndResident can be placed somewhere else - but it must follow the ROMTag and              ****
// **** it must not be placed in a different SECTION.                                             ****
// ***************************************************************************************************

extern struct InitTable InitTab;
extern APTR EndResident;  /* Below */

struct Resident __aligned ROMTag =     /* Do not change */
{
   RTC_MATCHWORD,
   &ROMTag,
   &EndResident,
   RTF_AUTOINIT,
   LIBRARYVERSION,
   NT_LIBRARY,
   0,
   &MyLibName[0],
   &MyLibID[0],
   &InitTab
};

APTR EndResident;

struct MyDataInit DataTab =            /* Do not change */
{
   INITBYTE(OFFSET(Node,        ln_Type),      NT_LIBRARY),
   0x80, (UBYTE)OFFSET(Node,    ln_Name),      (ULONG)&MyLibName[0],
   INITBYTE(OFFSET(Library,     lib_Flags),    LIBF_SUMUSED | LIBF_CHANGED),
   INITWORD(OFFSET(Library,     lib_Version),  LIBRARYVERSION),
   INITWORD(OFFSET(Library,     lib_Revision), LIBRARYREVISION),
   0x80, (UBYTE)OFFSET(Library, lib_IdString), (ULONG)&MyLibID[0],
   (ULONG)0
};


// ******************************************************************************************
// **** Function and Data Tables:                                                        ****
// ****                                                                                  ****
// **** The function and data tables have been placed here for traditional reasons.      ****
// **** Placing the RomTag structure before (-> UserLib.c) would also be a good idea,    ****
// **** but it depends on whether you would like to keep the "version" stuff separately. ****
// ******************************************************************************************

extern APTR FuncTab[];

// ***************************************************************
// **** Instead you may place ROMTag + Datatab directly, here ****
// **** (see UserLib.c). This may fix "Installer" version     ****
// **** checking problems, too.                               ****
// ***************************************************************

struct InitTable InitTab =             /* Do not change */
{
   (ULONG)              sizeof(struct LibraryBase),
   (APTR *)             &FuncTab[0],
   (struct MyDataInit *)&DataTab,
   (APTR)               InitLib
};

APTR FuncTab[] =
{
   OpenLib,
   CloseLib,
   ExpungeLib,
   ExtFuncLib,

   /* Add your own functions here                                   */
   /*                                                               */
   /* "libfunclist" is a text file which is automatically generated */
   /* by the FDToList utility from the .fd file of the library.     */

#include "libfunclist"

   (APTR)((LONG)-1)
};


// **************************************************************************************************
// **** InitLib():                                                                               ****
// ****                                                                                          ****
// **** This one is single-threaded by the Ramlib process. Theoretically you can do what         ****
// **** you like here, since you have full exclusive control over all the library code and data. ****
// **** But due to some bugs in Ramlib V37-40, you can easily cause a deadlock when opening      ****
// **** certain libraries here (which open other libraries, that open other libraries, that...). ****
// **************************************************************************************************

struct LibraryBase * __saveds __asm InitLib(register __a6 struct ExecBase *sysbase,
                                            register __a0 SEGLISTPTR seglist,
                                            register __d0 struct LibraryBase *libbase)
{
   //libbase->SysBase = sysbase;
   libbase->SegList = seglist;

   if (__UserLibInit(libbase) == 0L) return (libbase);

   __UserLibCleanup(libbase);

   {
      ULONG negsize, possize, fullsize;
      UBYTE *negptr = (UBYTE *)libbase;

      negsize  = libbase->Library.lib_NegSize;
      possize  = libbase->Library.lib_PosSize;
      fullsize = negsize + possize;
      negptr  -= negsize;

      FreeMem(negptr,fullsize);
   }

   return (NULL);
}

// **********************************************************************************************
// **** OpenLib():                                                                           ****
// ****                                                                                      ****
// **** This one is enclosed within a Forbid/Permit pair by Exec V37-40. Since a Wait() call ****
// **** would break this Forbid/Permit(), you are not allowed to start any operations that   ****
// **** may cause a Wait() during their processing. It's possible, that future OS versions   ****
// **** won't turn the multi-tasking off, but instead use semaphore protection for this      ****
// **** function.                                                                            ****
// ****                                                                                      ****
// **** Currently you only can bypass this restriction by supplying your own semaphore       ****
// **** mechanism.                                                                           ****
// **********************************************************************************************

struct LibraryBase * __saveds __asm OpenLib(register __a6 struct LibraryBase *libbase)
{
   libbase->Library.lib_OpenCnt += 1;

   libbase->Library.lib_Flags &= ~LIBF_DELEXP;

   return (libbase);
}

// **********************************************************************************************
// **** CloseLib():                                                                          ****
// ****                                                                                      ****
// **** This one is enclosed within a Forbid/Permit pair by Exec V37-40. Since a Wait() call ****
// **** would break this Forbid/Permit(), you are not allowed to start any operations that   ****
// **** may cause a Wait() during their processing. It's possible, that future OS versions   ****
// **** won't turn the multi-tasking off, but instead use semaphore protection for this      ****
// **** function.                                                                            ****
// ****                                                                                      ****
// **** Currently you only can bypass this restriction by supplying your own semaphore       ****
// **** mechanism.                                                                           ****
// **********************************************************************************************

SEGLISTPTR __saveds __asm CloseLib(register __a6 struct LibraryBase *libbase)
{
   libbase->Library.lib_OpenCnt -= 1;

   if (!libbase->Library.lib_OpenCnt)
   {
      if (libbase->Library.lib_Flags & LIBF_DELEXP)
      {
         return (ExpungeLib(libbase));
      }
   }

   return (NULL);
}

// **********************************************************************************************
// **** ExpungeLib():                                                                        ****
// ****                                                                                      ****
// **** This one is enclosed within a Forbid/Permit pair by Exec V37-40. Since a Wait() call ****
// **** would break this Forbid/Permit(), you are not allowed to start any operations that   ****
// **** may cause a Wait() during their processing. It's possible, that future OS versions   ****
// **** won't turn the multi-tasking off, but instead use semaphore protection for this      ****
// **** function.                                                                            ****
// ****                                                                                      ****
// **** Currently you only could bypass this restriction by supplying your own semaphore     ****
// **** mechanism - but since expunging can't be done twice, one should avoid it here.       ****
// **********************************************************************************************

SEGLISTPTR __saveds __asm ExpungeLib(register __a6 struct LibraryBase *libbase)
{
   SEGLISTPTR seglist;

   if (!libbase->Library.lib_OpenCnt)
   {
      if (__UserLibCleanup(libbase))
      {
         ULONG negsize, possize, fullsize;
         UBYTE *negptr = (UBYTE *)libbase;

         seglist = libbase->SegList;

         Remove((struct Node *)libbase);

         negsize  = libbase->Library.lib_NegSize;
         possize  = libbase->Library.lib_PosSize;
         fullsize = negsize + possize;
         negptr  -= negsize;

         FreeMem(negptr,fullsize);

         return (seglist);
      }
   }

   libbase->Library.lib_Flags |= LIBF_DELEXP;

   return (NULL);
}

// **********************************************************************************************
// **** ExtFunc():                                                                           ****
// ****                                                                                      ****
// **** This one is enclosed within a Forbid/Permit pair by Exec V37-40. Since a Wait() call ****
// **** would break this Forbid/Permit(), you are not allowed to start any operations that   ****
// **** may cause a Wait() during their processing. It's possible, that future OS versions   ****
// **** won't turn the multi-tasking off, but instead use semaphore protection for this      ****
// **** function.                                                                            ****
// ****                                                                                      ****
// **** Currently you only can bypass this restriction by supplying your own semaphore       ****
// **** mechanism - but since this function currently is unused, you should not touch        ****
// **** it, either.                                                                          ****
// **********************************************************************************************

ULONG __asm ExtFuncLib(void)
{
   return (NULL);
}


// ****************************************************************************************
// **** __SASC stuff:                                                                  ****
// ****                                                                                ****
// **** This is only for SAS/C - its intention is to turn off internal CTRL-C handling ****
// **** for standard C function and to avoid calls to exit() et al.                    ****
// ****************************************************************************************

#ifdef __SASC

#ifdef ARK_OLD_STDIO_FIX

ULONG XCEXIT       = NULL;  /* These symbols may be referenced by    */
ULONG _XCEXIT      = NULL;  /* some functions of sc.lib, but should  */
ULONG ONBREAK      = NULL;  /* never be used inside a shared library */
ULONG _ONBREAK     = NULL;
ULONG base         = NULL;
ULONG _base        = NULL;
ULONG ProgramName  = NULL;
ULONG _ProgramName = NULL;
ULONG StackPtr     = NULL;
ULONG _StackPtr    = NULL;
ULONG oserr        = NULL;
ULONG _oserr       = NULL;
ULONG OSERR        = NULL;
ULONG _OSERR       = NULL;

#endif  /* ARK_OLD_STDIO_FIX */

void __regargs __chkabort(void) { }  /* A shared library cannot be    */
void __regargs _CXBRK(void)     { }  /* CTRL-C aborted when doing I/O */

#endif  /* __SASC */


