

/*  UserLib.c  $ 31/03/98 MT $                                     */
/*                                                                 */
/*  Library initializers and functions to be called by LibStart.c  */
/*  Adapted from a source code by Andreas R. Kleinert              */


#include "exec/types.h"
#include "exec/memory.h"
#include "exec/libraries.h"
#include "exec/execbase.h"
#include "clib/exec_protos.h"
#include "pragmas/exec_sysbase_pragmas.h"
#include "librarybase.h"


#define LIBVERSIONREQUIRED 37


ULONG __saveds __asm __UserLibInit(register __a0 struct LibraryBase *);
ULONG __saveds __asm __UserLibCleanup(register __a0 struct LibraryBase *);

#ifdef CLASSLIBRARY
extern Class * __asm OpenLibraryClass(void);
extern ULONG __asm CloseLibraryClass(register __a0 Class *);
#endif

struct ExecBase      *SysBase       = NULL;
struct IntuitionBase *IntuitionBase = NULL;
struct GfxBase       *GfxBase       = NULL;
struct Library       *UtilityBase   = NULL;
struct LibraryBase   *ExternalBase  = NULL;


// *******************************************************************************************
// **** __UserLibInit():                                                                  ****
// ****                                                                                   ****
// **** Since this one is called by InitLib(), libraries not shareable between Processes  ****
// **** or libraries messing with RamLib (deadlock and crash) may not be opened here.     ****
// ****                                                                                   ****
// **** You may bypass this by calling this function from outside LibOpen(), but then you ****
// **** will have to a) protect it with a semaphore and b) make sure that libraries are   ****
// **** opened only once (when using global library bases).                               ****
// *******************************************************************************************

ULONG __saveds __asm __UserLibInit(register __a0 struct LibraryBase *library)
{
   SysBase = (*((struct ExecBase **)4L));

   IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library",LIBVERSIONREQUIRED);
   if (!IntuitionBase) return (1L);

   GfxBase = (struct GfxBase *)OpenLibrary("graphics.library",LIBVERSIONREQUIRED);
   if (!GfxBase) return (1L);

   UtilityBase = OpenLibrary("utility.library",LIBVERSIONREQUIRED);
   if (!UtilityBase) return (1L);

#ifdef CLASSLIBRARY
   if (!(library->Class = OpenLibraryClass()))
   {
      return (1L);
   }
#endif

   ExternalBase = library;

   //library->IntuitionBase = IntuitionBase;
   //library->GfxBase       = GfxBase;
   //library->SysBase       = SysBase;

   return (0L);
}

// ***************************************************************************************************
// **** __UserLibCleanup():                                                                       ****
// ****                                                                                           ****
// **** This one by default is called by ExpungeLib(), which only can take place once             ****
// **** and thus per definition is single-threaded.                                               ****
// ****                                                                                           ****
// **** When calling this from outside LibClose() instead, you will have to protect it with a     ****
// **** semaphore, since you don't know whether a given CloseLibrary(FooBase) may cause a Wait(). ****
// **** Additionally, there should be protection so that a library won't be closed twice.         ****
// ***************************************************************************************************

ULONG __saveds __asm __UserLibCleanup(register __a0 struct LibraryBase *library)
{
#ifdef CLASSLIBRARY
   if (library->Class)
   {
      if (!CloseLibraryClass(library->Class)) return (FALSE);
   }
#endif

   CloseLibrary(UtilityBase);
   CloseLibrary((struct Library *)GfxBase);
   CloseLibrary((struct Library *)IntuitionBase);

   return (TRUE);
}


