

#ifndef BOOPSISUPP_H
#define BOOPSISUPP_H


/*  Boopsi Support.h  $ 11/04/98 MT $  */


#ifndef EXEC_TYPES_H
#include "exec/types.h"
#endif

#ifndef INTUITION_CLASSES_H
#include "intuition/classes.h"
#endif


/* New types, useful for BOOPSI coding */

// *********************************************************
// **** Pointers to hook function and BOOPSI dispatcher ****
// *********************************************************

typedef ULONG __asm (*HOOKFUNCTION)(register __a0 struct Hook *, register __a2 APTR, register __a1 APTR);
typedef ULONG __asm (*DISPATCHER)(register __a0 Class *, register __a2 Object *, register __a1 Msg);


/* A number of fast inlined functions useful for BOOPSI coding: */
/* some of them override the equivalent ones in amiga.lib       */

// *****************************************************
// **** Invoke a method of an object's "true class" ****
// *****************************************************

ULONG __inline DoMethodA(Object *obj, Msg msg)
{
   return (((DISPATCHER)(OCLASS(obj)->cl_Dispatcher.h_Entry))(OCLASS(obj),obj,msg));
}

ULONG __inline DoMethod(Object *obj, ULONG method, ...)
{
   return (((DISPATCHER)(OCLASS(obj)->cl_Dispatcher.h_Entry))(OCLASS(obj),obj,(APTR)&method));
}

// *********************************************************************************
// **** Invoke a method of an object as if it did belong to the specified class ****
// *********************************************************************************

ULONG __inline CoerceMethodA(Class *class, Object *obj, Msg msg)
{
   return (((DISPATCHER)(class->cl_Dispatcher.h_Entry))(class,obj,msg));
}

ULONG __inline CoerceMethod(Class *class, Object *obj, ULONG method, ...)
{
   return (((DISPATCHER)(class->cl_Dispatcher.h_Entry))(class,obj,(APTR)&method));
}

// **********************************************************
// **** Invoke a method of an object as if it did belong ****
// **** to the superclass of the specified class         ****
// **********************************************************

ULONG __inline DoSuperMethodA(Class *class, Object *obj, Msg msg)
{
   return (((DISPATCHER)(class->cl_Super->cl_Dispatcher.h_Entry))(class->cl_Super,obj,msg));
}

ULONG __inline DoSuperMethod(Class *class, Object *obj, ULONG method, ...)
{
   return (((DISPATCHER)(class->cl_Super->cl_Dispatcher.h_Entry))(class->cl_Super,obj,(APTR)&method));
}

// *******************************************************************
// **** We can use this instead of DoMethodA() or CoerceMethodA() ****
// **** when we aren't sure this is a BOOPSI object               ****
// *******************************************************************

ULONG __inline CallHookFuncA(struct Hook *hook, APTR obj, APTR commandblock)
{
   return (((HOOKFUNCTION)(hook->h_Entry))(hook,obj,commandblock));
}

ULONG __inline CallHookFunc(struct Hook *hook, APTR obj, ULONG command, ...)
{
   return (((HOOKFUNCTION)(hook->h_Entry))(hook,obj,(APTR)&command));
}

// ********************************************************
// **** Notify an object of a change in its attributes ****
// ********************************************************

void __inline NotifyAttrsA(Object *obj, struct GadgetInfo *ginfo, ULONG flags, struct TagItem *attrlist)
{
   DoMethod(obj,OM_NOTIFY,attrlist,ginfo,flags);
}

void __inline NotifyAttrs(Object *obj, struct GadgetInfo *ginfo, ULONG flags, Tag attr1, ...)
{
   DoMethod(obj,OM_NOTIFY,&attr1,ginfo,flags);
}

// **************************************************************
// **** Ask an object to change the value of some attributes ****
// **************************************************************

void __inline UpdateAttrsA(Object *obj, struct GadgetInfo *ginfo, ULONG flags, struct TagItem *attrlist)
{
   DoMethod(obj,OM_UPDATE,attrlist,ginfo,flags);
}

void __inline UpdateAttrs(Object *obj, struct GadgetInfo *ginfo, ULONG flags, Tag attr1, ...)
{
   DoMethod(obj,OM_UPDATE,&attr1,ginfo,flags);
}


#endif


