

/*  Fast Layout Library  $ 31/05/95 MT $  */


#include "exec/types.h"
#include "exec/libraries.h"
#include "exec/memory.h"
#include "graphics/gfx.h"
#include "graphics/rastport.h"
#include "intuition/intuition.h"
#include "intuition/gadgetclass.h"
#include "libraries/gadtools.h"
#include "utility/tagitem.h"
#include "proto/exec.h"
#include "proto/graphics.h"
#include "proto/intuition.h"
#include "proto/gadtools.h"
#include "proto/utility.h"
#include "rl_fastlayout.h"
#include "fastlayout.h"
#include "fl_private.h"


static ULONG TextWidth(struct TextAttr *, UBYTE *);

static struct TextAttr FLTextAttr = { "topaz.font", 8, 0, 0 };

struct TextAttr *FL_Private1 = &FLTextAttr;
ULONG (*FL_Private2)() = TextWidth;


/*
struct RastPort *CreateRastPort(WORD width, WORD height, WORD depth)
{
   struct RastPort *rp;
   struct BitMap *bm;
   UBYTE *planes;
   BYTE a;
   ULONG size, psize;

   size = (((width + 15) / 16) * 2) * height * depth;
   psize = size / depth;

   if (planes = AllocVec(size,MEMF_CHIP|MEMF_CLEAR))
   {
      if (bm = AllocVec(sizeof(struct BitMap),MEMF_CLEAR))
      {
         if (rp = AllocVec(sizeof(struct RastPort),MEMF_CLEAR))
         {
            InitBitMap(bm,depth,width,height);
            for (a = 0; a < depth; a++) bm->Planes[a] = planes + a * psize;
            InitRastPort(rp);
            rp->BitMap = bm;

            return (rp);
         }

         FreeVec(bm);
      }

      FreeVec(planes);
   }

   return (NULL);
}

void DeleteRastPort(struct RastPort *rp)
{
   if (rp)
   {
      if (rp->BitMap)
      {
         if (rp->BitMap->Planes[0]) FreeVec(rp->BitMap->Planes[0]);
         FreeVec(rp->BitMap);
      }

      FreeVec(rp);
   }
}
*/

#ifdef RESLIB
static void GetTextAttr(struct TextAttr **ta, struct LayoutItem *item, struct Layout *layout)
#else
static void __asm GetTextAttr(register __a0 struct TextAttr **ta, register __a1 struct LayoutItem *item, register __a2 struct Layout *layout)
#endif
{
   if (ta) if (!(*ta = item->TextAttr)) *ta = (layout? layout->TextAttr: &FLTextAttr);
}

static ULONG TextWidth(struct TextAttr *ta, UBYTE *testo)
{
   struct IntuiText it = { 0 };
   struct TextFont *tf;
   ULONG width = 0L;

   if (testo && (tf = OpenFont(ta)))
   {
      it.ITextFont = ta;
      it.IText = testo;
      width = IntuiTextLength(&it);

      /* Sporco trucco per tenere conto dell'underscore */

      if (strchr(testo,'_'))
      {
         it.IText = "_";
         width -= IntuiTextLength(&it);
      }

      CloseFont(tf);
   }

   return (width);
}

#ifdef RESLIB
static ULONG FontWidth(struct TextAttr *ta)
#else
static ULONG __asm FontWidth(register __a0 struct TextAttr *ta)
#endif
{
   struct TextFont *tf;
   UWORD fontwidth = 8;

   if (tf = OpenFont(ta))
   {
      fontwidth = tf->tf_XSize;
      CloseFont(tf);
   }

   return (fontwidth);
}

/****** fastlayout.library/BuildLItemListA **********************************
*
*   NAME
*       BuildLItemListA -- Builds a linked list of LayoutItem structures.
*       BuildLItemList -- Varargs version of BuildLItemListA().
*
*   SYNOPSIS
*       first = BuildLItemListA(includelists, taglist)
*       D0                      D0              A0
*
*       APTR BuildLItemListA(ULONG, struct TagItem *);
*
*       first = BuildLItemList(includelists, firsttag, ...)
*
*       APTR BuildLItemList(ULONG, Tag, ...);
*
*   FUNCTION
*       This function takes the LayoutItem structures whose pointers you
*       pass with the tags and links them into a list.
*       If you pass TRUE for includelist, when a structure is already part
*       of a list the entire list starting with it is linked into the new
*       list, otherwise only the single item is linked, and the link with the
*       subsequent list is severed.
*       Either way, if there is a list before the item, the link with this
*       list will be severed before inserting the item into the new list.
*       Furthermore, if an item has a sub-tree of LayoutItems this sub-tree
*       will be preserved.
*       Take care using this function in order to always build correct lists
*       and not to lose any pointer to items or sub-lists, as in this case
*       you wouldn't be able to free them later.
*
*   INPUTS
*       includelists - specify FALSE if you want to link only the single
*                      items into the new list, or TRUE if you want to
*                      include also any list of which the items are the head.
*       taglist - pointer to an array of TagItem pairs. The only allowed
*                 tag is FL_Item, used to pass a pointer to a LayoutItem
*                 structure.
*
*   RESULT
*       first - pointer to the first item (head) of the linked list, or
*               NULL if no items were passed.
*
*   EXAMPLE
*
*       \* Let's build a list from single items only: *\
*
*       it1 = CreateLayoutItem(IT_ZONE,LT_HORIZ,0,FL_AutoSpace,TRUE,TAG_END);
*       it2 = CreateLayoutItem(IT_ZONE,LT_VERT,0,FL_AutoSpace,TRUE,TAG_END);
*       first = BuildLItemList(FALSE,FL_Item,it1,FL_Item,it2,TAG_END);
*
*       \* Now the list contains it1 and it2. *\
*
*       \* Now let's build a new list from our old list and a new item: *\
*
*       it3 = CreateLayoutItem(IT_ZONE,LT_HORIZ,0,FL_AutoSpace,TRUE,TAG_END);
*       first = BuildLItemList(TRUE,FL_Item,first,FL_Item,it3,TAG_END);
*
*       \* Now the list contains it1, it2 and it3. *\
*
*       \* Finally, let's build a new list from the first item of our *\
*       \* existing list and yet another new item:                    *\
*
*       it4 = CreateLayoutItem(IT_ZONE,LT_HORIZ,0,FL_AutoSpace,TRUE,TAG_END);
*       first = BuildLItemList(FALSE,FL_Item,first,FL_Item,it4,TAG_END);
*
*       \* Now the list contains it1 and it4.        *\
*       \* The other two items form a separate list. *\
*
*   CAUTION
*       You must be VERY CAREFUL not to pass the same item or sub-list
*       twice, as this would cause the function to fall into an endless
*       loop!
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       SetLayoutItemAttrsA()
*
****************************************************************************/

struct LayoutItem * ASM BuildLItemListA(R_D0 ULONG includelists, R_A0 struct TagItem *taglist)
{
   struct TagItem *ti;
#ifdef RESLIB
   struct TagItem *tptr = taglist;
#endif
   struct LayoutItem *li0, *li = NULL, *first = NULL;

#ifdef RESLIB
   while (ti = NextTagItem(&tptr))
#else
   while (ti = NextTagItem(&taglist))
#endif
   {
      if (ti->ti_Tag == FL_Item)
      {
         if (!(li0 = (struct LayoutItem *)(ti->ti_Data)))
         {
            continue;
         }

         if (li0->Prev) li0->Prev->Next = NULL;

         if (!first)
         {
            first = li0;
            first->Prev = NULL;
         }
         else
         {
            li->Next = li0;
            li0->Prev = li;
         }

         if (!includelists)
         {
            li = li0;
            if (li->Next) li->Next->Prev = NULL;
            li->Next = NULL;
         }
         else
         {
            for (li = li0; li->Next; li = li->Next) ;
         }
      }
   }

   return (first);
}

#ifndef RESLIB

struct LayoutItem *BuildLItemList(ULONG includelists, Tag tag1, ...)
{
   return (BuildLItemListA(includelists,(struct TagItem *)&tag1));
}

#endif

/****** fastlayout.library/FreeLayoutItems **********************************
*
*   NAME
*       FreeLayoutItems -- Deallocates one or more LayoutItem structures.
*
*   SYNOPSIS
*       FreeLayoutItems(root)
*                       A0
*
*       void FreeLayoutItems(APTR);
*
*   FUNCTION
*       Deallocates the entire tree of LayoutItem structures of which the
*       passed LayoutItem is the root.
*
*   INPUTS
*       root - pointer to the root of the LayoutItem tree to be freed.
*              It is safe to call this function with a NULL parameter.
*
*   NOTES
*       It is dangerous to free a sub-tree of LayoutItem if there is a
*       LayoutItem that references it. You should only free the entire
*       tree which you created, and not parts of it.
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       CreateLayoutItemA(), FreeLayout()
*
****************************************************************************/

void ASM FreeLayoutItems(R_A0 struct LayoutItem *first)
{
   struct LayoutItem *li;

   while (first)
   {
      if (first->First)
      {
         FreeLayoutItems(first->First);
      }

      li = first;
      first = first->Next;
      FreeVec(li);
   }
}

/****** fastlayout.library/FreeLayout ***************************************
*
*   NAME
*       FreeLayout -- Deallocates a Layout structure.
*
*   SYNOPSIS
*       FreeLayout(layout)
*                  A0
*
*       void FreeLayout(APTR);
*
*   FUNCTION
*       Deallocates the specified Layout structure.
*
*   INPUTS
*       layout - pointer to the Layout structure to be freed.
*                It is safe to call this function with a NULL parameter.
*
*   NOTES
*       You should only call this function at your program's exit, as there
*       is no need to deallocate a Layout structure if you only want to
*       change it in some way.
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       CreateLayoutA(), CreateLayoutItemA()
*
****************************************************************************/

void ASM FreeLayout(R_A0 struct Layout *layout)
{
   if (layout)
   {
      FreeVec(layout);
   }
}

/****** fastlayout.library/SetLayoutAttrsA **********************************
*
*   NAME
*       SetLayoutAttrsA -- Sets or modifies the attributes of a Layout.
*       SetLayoutAttrs -- Varargs version of SetLayoutAttrsA().
*
*   SYNOPSIS
*       success = SetLayoutAttrsA(layout, taglist)
*       D0                        A0      A1
*
*       ULONG SetLayoutAttrsA(APTR, struct TagItem *);
*
*       success = SetLayoutAttrs(layout, firsttag, ...)
*
*       ULONG SetLayoutAttrs(APTR, Tag, ...);
*
*   FUNCTION
*       With this function you can modify the attributes of a Layout.
*       The values of these attributes are the default values for the
*       corresponding attributes of the LayoutItems of a particular tree,
*       when these are left undefined.
*       The tags you pass to this function are exactly the same you pass
*       to the CreateLayoutA() function, and have the same meaning.
*
*   INPUTS
*       layout - pointer to the Layout structure whose attributes are to
*                be modified.
*       taglist - pointer to an array of TagItem pairs.
*
*   TAGS
*       The tags currently accepted by this function are:
*
*       FL_TextAttr (struct TextAttr *) - Pointer to a valid TextAttr
*                                         structure, used as the default
*                                         TextAttr for LayoutItems which
*                                         haven't the FL_TextAttr tag.
*                                         If you pass NULL, the default
*                                         font (topaz 8) will be used.
*       FL_FrameWidth (UWORD) - Width of the frame of the gadgets in this
*                               layout, to be used as the default value for
*                               LayoutItems without this tag. This is the
*                               width of ONE of the two (left and right)
*                               vertical borders of the gadget. Note that
*                               this tag is only valid for AutoSize gadgets.
*                               If you pass -1, the default (19 pixels) will
*                               be used.
*       FL_FrameHeight (UWORD) - Height of the frame of the gadgets in this
*                                layout, to be used as the default value for
*                                LayoutItems without this tag. This is the
*                                height of ONE of the two (top and bottom)
*                                horizontal borders of the gadget. Note that
*                                this tag is only valid for AutoSize gadgets.
*                                If you pass -1, the default (3 pixels) will
*                                be used.
*       FL_BorderWidth (UWORD) - Width of the border of the items in this
*                                layout, to be used as the default value for
*                                LayoutItems with a border and without this
*                                tag. This refers to borders added by the
*                                library, not to gadget frames.
*                                If you pass -1, the default (8 pixels) will
*                                be used.
*       FL_BorderHeight (UWORD) - Height of the border of the items in this
*                                 layout, to be used as the default value for
*                                 LayoutItems with a border and without this
*                                 tag. This refers to borders added by the
*                                 library, not to gadget frames.
*                                 If you pass -1, the default (5 pixels) will
*                                 be used.
*
*   RESULT
*       success - TRUE if all went ok, FALSE if the new attributes could not
*                 be set for some reasons.
*
*   EXAMPLE
*
*       \* Changes the default border width of items *\
*
*       ok = SetLayoutAttrs(layout,FL_BorderWidth,12,TAG_END);
*
*   NOTES
*       If a tag is repeated, only the last instance will be taken into
*       consideration. If two or more conflicting tags are passed, only the
*       last one will be taken into consideration.
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       GetLayoutAttrsA(), CreateLayoutA(), SetLayoutItemAttrsA()
*
****************************************************************************/

ULONG ASM SetLayoutAttrsA(R_A0 struct Layout *layout, R_A1 struct TagItem *taglist)
{
   struct TagItem *ti;
#ifdef RESLIB
   struct TagItem *tptr = taglist;
#endif

   if (!layout) return (FALSE);

#ifdef RESLIB
   while (ti = NextTagItem(&tptr))
#else
   while (ti = NextTagItem(&taglist))
#endif
   {
      switch (ti->ti_Tag)
      {
         case FL_TextAttr:
            layout->TextAttr = (struct TextAttr *)(ti->ti_Data);
            break;

         case FL_BorderWidth:
            layout->BorderSizes[0] = ti->ti_Data;
            break;

         case FL_BorderHeight:
            layout->BorderSizes[1] = ti->ti_Data;
            break;

         case FL_FrameWidth:
            layout->GadgetHDelta = ti->ti_Data + ti->ti_Data;
            break;

         case FL_FrameHeight:
            layout->GadgetVDelta = ti->ti_Data + ti->ti_Data;
            break;
      }
   }

   if (!layout->TextAttr) layout->TextAttr = &FLTextAttr;
   if (layout->BorderSizes[0] < 0) layout->BorderSizes[0] = DEFBHSIZE;
   if (layout->BorderSizes[1] < 0) layout->BorderSizes[1] = DEFBVSIZE;
   if (layout->GadgetHDelta & 0x8000) layout->GadgetHDelta = DEFGADGHDELTA;
   if (layout->GadgetVDelta & 0x8000) layout->GadgetVDelta = DEFGADGVDELTA;

   return (TRUE);
}

#ifndef RESLIB

ULONG SetLayoutAttrs(struct Layout *layout, Tag tag1, ...)
{
   return (SetLayoutAttrsA(layout,(struct TagItem *)&tag1));
}

#endif

/****** fastlayout.library/CreateLayoutA ************************************
*
*   NAME
*       CreateLayoutA -- Allocates a new Layout structure.
*       CreateLayout -- Varargs version of CreateLayoutA().
*
*   SYNOPSIS
*       layout = CreateLayoutA(layouttype, taglist)
*       D0                     D0          A0
*
*       APTR CreateLayoutA(ULONG, struct TagItem *);
*
*       layout = CreateLayout(layouttype, firsttag, ...)
*
*       APTR CreateLayout(ULONG, Tag, ...);
*
*   FUNCTION
*       This function allocates a new Layout structure, used as a common
*       context reference by all the functions that operate on a
*       LayoutItem tree. Note that the format of the Layout structure is
*       library private, so this function will return an APTR pointer.
*
*   INPUTS
*       layouttype - type (direction) of the layout of the root LayoutItem.
*                    This can be LT_HORIZ or LT_VERT.
*       taglist - pointer to an array of TagItem pairs.
*
*   TAGS
*       The tags currently accepted by this function are:
*
*       FL_TextAttr (struct TextAttr *) - Pointer to a valid TextAttr
*                                         structure, used as the default
*                                         TextAttr for LayoutItems which
*                                         haven't the FL_TextAttr tag.
*                                         The default is topaz 8.
*       FL_FrameWidth (UWORD) - Width of the frame of the gadgets in this
*                               layout, to be used as the default value for
*                               LayoutItems without this tag. This is the
*                               width of ONE of the two (left and right)
*                               vertical borders of the gadget. Note that
*                               this tag is only valid for AutoSize gadgets.
*                               The default is 19 pixels.
*       FL_FrameHeight (UWORD) - Height of the frame of the gadgets in this
*                                layout, to be used as the default value for
*                                LayoutItems without this tag. This is the
*                                height of ONE of the two (top and bottom)
*                                horizontal borders of the gadget. Note that
*                                this tag is only valid for AutoSize gadgets.
*                                The default is 3 pixels.
*       FL_BorderWidth (UWORD) - Width of the border of the items in this
*                                layout, to be used as the default value for
*                                LayoutItems with a border and without this
*                                tag. This refers to borders added by the
*                                library, not to gadget frames.
*                                The default is 8 pixels.
*       FL_BorderHeight (UWORD) - Height of the border of the items in this
*                                 layout, to be used as the default value for
*                                 LayoutItems with a border and without this
*                                 tag. This refers to borders added by the
*                                 library, not to gadget frames.
*                                 The default is 5 pixels.
*
*   RESULT
*       layout - pointer to a valid Layout structure.
*
*   EXAMPLE
*
*       layout = CreateLayout(LT_HORIZ,FL_TextAttr,&textattr,TAG_END);
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       FreeLayout(), SetLayoutAttrsA(), CreateLayoutItemA()
*
****************************************************************************/

struct Layout * ASM CreateLayoutA(R_D0 ULONG ltype, R_A0 struct TagItem *taglist)
{
   struct Layout *layout;

   if (layout = AllocVec(sizeof(struct Layout),MEMF_CLEAR))
   {
      layout->LayoutType = ltype;
      layout->TextAttr = &FLTextAttr;
      layout->GadgetHDelta = DEFGADGHDELTA;
      layout->GadgetVDelta = DEFGADGVDELTA;
      layout->BorderSizes[0] = DEFBHSIZE;
      layout->BorderSizes[1] = DEFBVSIZE;
      SetLayoutAttrsA(layout,taglist);
   }

   return (layout);
}

#ifndef RESLIB

struct Layout *CreateLayout(ULONG ltype, Tag tag1, ...)
{
   return (CreateLayoutA(ltype,(struct TagItem *)&tag1));
}

#endif

#ifdef RESLIB
static struct LayoutItem *FindLayoutItemParent(struct LayoutItem *parent, struct LayoutItem *root, ULONG type, ULONG id)
#else
static struct LayoutItem * __asm FindLayoutItemParent(register __a0 struct LayoutItem *parent, register __a1 struct LayoutItem *root, register __d0 ULONG type, register __d1 ULONG id)
#endif
{
   struct LayoutItem *li;

   while (root)
   {
      if ((root->Type == type) && (root->Id == id)) return (parent);
      if (li = FindLayoutItemParent(root,root->First,type,id)) return (li);

      root = root->Next;
   }

   return (NULL);
}

#ifdef RESLIB
static struct LayoutItem *FindIndicator(struct LayoutItem *root, ULONG indicatorid)
#else
static struct LayoutItem * __asm FindIndicator(register __a0 struct LayoutItem *root, register __d0 ULONG indicatorid)
#endif
{
   struct LayoutItem *li;

   while (root)
   {
      if ((root->Type == IT_GADGET) && (root->IndicatorId == indicatorid)) return (root);
      if (li = FindIndicator(root->First,indicatorid)) return (li);

      root = root->Next;
   }

   return (NULL);
}

/****** fastlayout.library/FindLayoutItem ***********************************
*
*   NAME
*       FindLayoutItem -- Searches for a specified LayoutItem in a tree.
*
*   SYNOPSIS
*       item = FindLayoutItem(root, type, id)
*       D0                    A0    D0    D1
*
*       APTR FindLayoutItem(APTR, ULONG, ULONG);
*
*   FUNCTION
*       This function returns the pointer of a LayoutItem with the specified
*       type and id if there is such a LayoutItem in the tree whose pointer
*       you pass to the function, else it returns NULL.
*
*   INPUTS
*       root - pointer to the root of the LayoutItem tree.
*       type - type of the required LayoutItem (IT_ZONE, IT_GADGET or
*              IT_SPACE). This is required as LayoutItems of different types
*              can have the same id (although you usually will search for a
*              LayoutItem of type IT_GADGET).
*       id - id of the required LayoutItem, as specified in the
*            CreateLayoutItemA() function call.
*
*   RESULT
*       item - pointer to the required item, or NULL if it wasn't found.
*
*   EXAMPLE
*
*       gadg4_item = FindLayoutItem(root,IT_GADGET,4);
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       CreateLayoutItemA()
*
****************************************************************************/

struct LayoutItem * ASM FindLayoutItem(R_A0 struct LayoutItem *root, R_D0 ULONG type, R_D1 ULONG id)
{
   struct LayoutItem *li;

   while (root)
   {
      if ((root->Type == type) && (root->Id == id)) return (root);
      if (li = FindLayoutItem(root->First,type,id)) return (li);

      root = root->Next;
   }

   return (NULL);
}

/****** fastlayout.library/SetLayoutItemAttrsA ******************************
*
*   NAME
*       SetLayoutItemAttrsA -- Sets or modifies the attributes of an item.
*       SetLayoutItemAttrs -- Varargs version of SetLayoutItemAttrsA().
*
*   SYNOPSIS
*       firstchild = SetLayoutItemAttrsA(layout, layoutitem, taglist)
*       D0                               A0      A1          A2
*
*       APTR SetLayoutItemAttrsA(APTR, APTR, struct TagItem *);
*
*       firstchild = SetLayoutItemAttrs(layout, layoutitem, firsttag, ...)
*
*       APTR SetLayoutItemAttrs(APTR, APTR, Tag, ...);
*
*   FUNCTION
*       With this function you can modify the attributes of a LayoutItem.
*       The tags you pass to this function are exactly the same you pass
*       to the CreateLayoutItemA() function, and have the same meaning. If
*       you pass one or more FL_Item tags, the sub-tree of this LayoutItem is
*       replaced by the new one composed by the items you specified, else
*       the sub-items of this LayoutItem are left alone. In both cases the
*       function returns the pointer to the first sub-item this item had
*       before you called SetLayoutItemAttrsA().
*
*   INPUTS
*       layout - pointer to a reference Layout structure. As of V40, this
*                is not used, so just pass NULL here.
*       layoutitem - pointer to the LayoutItem whose attributes are to
*                    be modified.
*       taglist - pointer to an array of TagItem pairs.
*
*   TAGS
*       FL_AutoSpace (ULONG) - If this tag is present, the sub-items of this
*                              item (which must be a IT_ZONE or the tag will
*                              be ignored) will be automatically spaced.
*                              If there is at least one sub-item of variable
*                              size, the value of this tag is used as the
*                              fixed pixel number to place between sub-items.
*                              If all the sub-items have a fixed size, the
*                              spacing will be variable depending on the size
*                              of the entire layout, and the value of this
*                              tag will be the minimum size of the spacing.
*                              If you pass the value -1 (or ~0) the automatic
*                              spacing will be disabled.
*       FL_AlignItems (BOOL) - As of V40, this tag has only meaning for
*                              vertical zones (i.e. items of type IT_ZONE
*                              with a vertical layout), and its purpose is
*                              to align the left (or right) sides of the
*                              gadgets they contain. Currently (V40) this
*                              tag is only guaranteed to work for gadgets
*                              of the kinds BUTTON, STRING, INTEGER, TEXT,
*                              NUMBER, CYCLE, CHECKBOX, SCROLLER and SLIDER
*                              without level indicator.
*                              Not supported yet are the kinds PALETTE,
*                              LISTVIEW and SLIDER with level indicator.
*                              Furthermore, for this tag to work, the
*                              justification of the gadgets must be JT_FLUSH.
*       FL_Item (APTR) - Pointer to a valid LayoutItem structure, which will
*                        become a child of this LayoutItem. You can pass more
*                        than one FL_Item tag. It is incorrect to pass a
*                        FL_Item tag if this is not a IT_ZONE. You can pass a
*                        NULL to detach the current sub-tree of this zone.
*       FL_Prev (APTR) - Pointer to a valid LayoutItem structure, which will
*                        become the previous item in this one's list. This is
*                        used to link items to form a list. However this is
*                        only necessary when this item is not a child of
*                        another one, because if it is, it will be
*                        automatically linked with the other children (if
*                        they exist). You can also pass a NULL to make this
*                        item the first one of its list.
*       FL_Next (APTR) - Pointer to a valid LayoutItem structure, which will
*                        become the next item in this one's list. This is
*                        used to link items to form a list. However this is
*                        only necessary when this item is not a child of
*                        another one, because if it is, it will be
*                        automatically linked with the other children (if
*                        they exist). You can also pass a NULL to make this
*                        item the last one of its list.
*       FL_GadgetText (STRPTR *) - Pointer to the pointer to the string this
*                                  item, when actualized into a gadget, will
*                                  contain. The double indirection is useful
*                                  to be able to change the string pointer
*                                  without having to reassign it to the
*                                  LayoutItem. It is incorrect to pass a
*                                  FL_GadgetText if this is not a IT_GADGET.
*                                  If you pass a NULL here, there should be
*                                  no text in the gadget, or its minimum size
*                                  will be calculated incorrectly.
*       FL_TextAttr (struct TextAttr *) - Pointer to a valid TextAttr
*                                         structure describing the font that
*                                         will be used for the gadget
*                                         associated with this LayoutItem.
*                                         If you pass a NULL here, the font
*                                         of the Layout structure will be
*                                         used instead. It is incorrect to
*                                         pass a FL_TextAttr tag if this is
*                                         not a IT_GADGET.
*       FL_FrameWidth (UWORD) - Width of the frame of the gadget associated
*                               to this item. This is the width of ONE of the
*                               two (left and right) vertical borders of the
*                               gadget. Note that this tag is only valid for
*                               AutoSize gadgets. If you pass -1 (or ~0) this
*                               item will use the default value specified for
*                               the Layout structure. It is incorrect to pass
*                               a FL_FrameWidth tag if this is not an
*                               IT_GADGET.
*       FL_FrameHeight (UWORD) - Height of the frame of the gadget associated
*                                to this item. This is the height of ONE of
*                                the two (top and bottom) horizontal borders
*                                of the gadget. Note that this tag is only
*                                valid for AutoSize gadgets. If you pass -1
*                                (or ~0) this item will use the default value
*                                specified for the Layout structure. It is
*                                incorrect to pass a FL_FrameHeight tag if
*                                this is not a IT_GADGET.
*       FL_BorderType (ULONG) - Type of border this item must have. This is
*                               used mainly for IT_ZONE items, as the gadgets
*                               already have a border. The layout will take
*                               the size of the border into account; to
*                               actually render the border you must call the
*                               RefreshLayout() function. The currently
*                               defined types are:
*
*                               BT_NONE - No border.
*                               BT_SPACE - Invisible border.
*                               BT_RAISED - The standard gadtools border.
*                               BT_RECESSED - As above, but recessed.
*                               BT_LINE - A furrow (recessed gadtools ridge).
*
*                               Note that there is no BT_RIDGE type yet. This
*                               could change in the future if gadtools will
*                               support a multi-line string-gadget without
*                               borders.
*       FL_BorderWidth (UWORD) - Width of the border of this item, if it has
*                                one. This refers to the border added by the
*                                library with FL_BorderType.
*                                If you pass -1, the default value of the
*                                Layout structure will be used.
*       FL_BorderHeight (UWORD) - Height of the border of this item, if it
*                                 has one. This refers to the border added by
*                                 the library with FL_BorderType.
*                                 If you pass -1, the default value of the
*                                 Layout structure will be used.
*       FL_Justify (ULONG) - Type of justification of this item in the
*                            secundary direction, that is the direction
*                            perpendicular to the one of the layout of the
*                            parent item. The types currently defined are:
*
*                            JT_FLUSH - Spread the item using all the
*                                       available space.
*                            JT_LEFTTOP - Place the item to the left (top)
*                                         side of its parent item.
*                            JT_RIGHTBOTTOM - Place the item to the right
*                                             (bottom) side of its parent
*                                             item.
*                            JT_CENTER - Place the item in the center of
*                                        its parent item.
*
*                            With the exception of the JT_FLUSH case, the
*                            item will be reduced to its minimum secundary
*                            size when justified.
*       FL_Size (ULONG) - Fixed absolute size in pixels of this item in the
*                         direction of the layout of its parent item. If
*                         you specify this tag, the item will always have
*                         the same size, i. e. won't adapt to the font (if
*                         a IT_GADGET) or to its contents (if a IT_ZONE).
*       FL_RelSize (ULONG) - Size of this item in the direction of its
*                            parent's layout, in relation to the space
*                            available (the size of the parent item minus
*                            the size of all fixed-size items). This is the
*                            "weight" of this item, so the actual size will
*                            depend on the relative size of the other items
*                            with the FL_RelSize tag. For example, if you
*                            want only two items, one of which has always a
*                            size equal to the half of the other one's, you
*                            can specify a FL_RelSize of 2 and 1 for the
*                            items; you could also specify 100 and 50.
*                            The actual size is also limited by the minimum
*                            size of the item, so when the layout is at its
*                            minimum size the FL_RelSize gadgets won't
*                            respect their relative weights (unless the
*                            minimum sizes, too, reflect them).
*       FL_AutoSize (ULONG) - If you specify this tag the item will assume
*                             a size (in the direction of its parent's
*                             layout) that is the minimum to allow its
*                             contents to fit into it. In other words, the
*                             IT_GADGET items will be sized taking into
*                             account their text, font, frame and border
*                             size, and other specific attributes; the
*                             IT_ZONE items will be sized taking into account
*                             the sub-items they contain. The IT_SPACE items
*                             will always be set to their minimum size as
*                             specified with the FL_MinSize tag.
*                             Note that to determine the size in the
*                             secundary direction you must use the FL_Justify
*                             tag and pass the appropriate tags to the
*                             parent(s) of this item.
*                             As of V40 the actual value passed with this tag
*                             is ignored, so you should pass a value of zero.
*       FL_MinSize (ULONG) - The absolute minimum size this item can reach.
*                            It is guaranteed that the item will never be
*                            smaller than this, no matter which other tags
*                            are used. This refers only to the direction of
*                            the layout of this item's parent; to determine
*                            the minimum size in the secundary direction you
*                            must set it in the parents themselves.
*       FL_Labels (STRPTR *) - If this item will become a list-type gadget,
*                              that is (currently) a ListView, a MX or a
*                              Cycle gadget, you should pass with this tag
*                              the pointer to the array of labels the gadget
*                              will display. This is necessary to correctly
*                              find out the minimum size of the item.
*                              The array must be NULL-terminated.
*                              Note: as of V40, this tag is not required for
*                              items which will become ListViews.
*       FL_ListInfo (struct ListInfo *) - If this item will become a
*                                         list-type gadget, that is
*                                         (currently) a ListView, a MX or a
*                                         Cycle gadget, you should pass with
*                                         this tag the pointer to a ListInfo
*                                         structure describing the attributes
*                                         of the gadget. The structure fields
*                                         (and their meaning) are:
*
*                                         GadgetKind - The exact kind of
*                                                      the gadget. It can
*                                                      be LISTVIEW_KIND,
*                                                      MX_KIND or CYCLE_KIND.
*
*                                         TextPlace - For MX gadgets this
*                                                     indicates where the
*                                                     labels must go with
*                                                     respect to the buttons.
*                                                     For the other two kinds
*                                                     this indicates the
*                                                     position of the gadget
*                                                     text. The values can be
*                                                     the usual PLACETEXT_#?
*                                                     definitions.
*
*                                         Spacing - The amount of spacing
*                                                   between labels of a
*                                                   MX or ListView gadget.
*                                                   This is ignored for
*                                                   Cycle gadgets.
*
*                                         GadgetWidth - For MX gadgets, this
*                                                       is the width of the
*                                                       buttons in pixels.
*                                                       If your program runs
*                                                       under an OS previous
*                                                       to 3.0, this value
*                                                       will be ignored and
*                                                       MX_WIDTH (17) will be
*                                                       assumed. If you put 0
*                                                       here, under 3.0+ an
*                                                       automatic value will
*                                                       be used (namely font
*                                                       height + 1).
*                                                       For ListView gadgets,
*                                                       this is the width of
*                                                       the scrollbar; if you
*                                                       put 0 here, you will
*                                                       get the default (16).
*                                                       This field is ignored
*                                                       for Cycle gadgets.
*
*                                         GadgetHeight - For MX gadgets, this
*                                                        is the height of the
*                                                        buttons in pixels.
*                                                        If your program runs
*                                                        under an OS previous
*                                                        to 3.0, this value
*                                                        will be ignored and
*                                                        MX_HEIGHT (9) will
*                                                        be assumed. If you
*                                                        put 0 here, under
*                                                        3.0+ an automatic
*                                                        value will be used
*                                                        (namely font height
*                                                        + 1).
*                                                        For ListView gadgets
*                                                        this is the height
*                                                        of the indicator
*                                                        placed under the
*                                                        gadget (if there is
*                                                        one). If you put
*                                                        0 here, it will be
*                                                        assumed an height
*                                                        equal to the font
*                                                        height plus the
*                                                        double of the value
*                                                        passed with the
*                                                        FL_FrameHeight tag.
*                                                        It is recommended
*                                                        that you always use
*                                                        this value when you
*                                                        create the string
*                                                        gadget to place
*                                                        under your ListView.
*                                                        This field is
*                                                        ignored for Cycle
*                                                        gadgets.
*
*                                         Flags - Currently (V40) only one
*                                                 flag is defined:
*                                                 LIF_ADJUSTTOP causes a
*                                                 ListView to be vertically
*                                                 centered inside its item.
*                                                 This is useful because
*                                                 GadTools may shorten the
*                                                 ListView to make it contain
*                                                 an integral number of
*                                                 labels, and this way the
*                                                 extra space gets equally
*                                                 distributed on both sides
*                                                 of the ListView.
*                                                 This flag is ignored for
*                                                 MX and Cycle gadgets.
*       FL_WidthGroup (ULONG) - If you want two or more items to have the
*                               same width, you can specify this tag for each
*                               of them, passing the same (arbitrary) number.
*                               This will form an ideal "class" of items
*                               having the same width.
*                               Note: currently this works only for items
*                               of type IT_GADGET with the FL_AutoSize tag.
*       FL_HeightGroup (ULONG) - If you want two or more items to have the
*                                same height, you can specify this tag for
*                                each of them, passing the same (arbitrary)
*                                number.
*                                This will form an ideal "class" of items
*                                having the same height.
*                                Note: currently this works only for items
*                                of type IT_GADGET with the FL_AutoSize tag.
*       FL_SizeGroup (ULONG) - If you want two or more items to have the
*                              same size, you can specify this tag for each
*                              of them, passing the same (arbitrary) number.
*                              This will form an ideal "class" of items
*                              having the same size. This is equivalent to
*                              passing both the FL_WidthGroup and
*                              FL_HeightGroup tags with the same number.
*                              Note: currently this works only for items
*                              of type IT_GADGET with the FL_AutoSize tag.
*       FL_HorizLevel (ULONG) - If this item will become a Slider with its
*                               level displayed at the gadget's left or right
*                               you should specify with this tag the maximum
*                               length of the level string to be displayed.
*                               This is the same value you would pass with
*                               the gadtools GTSL_MaxLevelLen tag. This is
*                               necessary to correctly find out the minimum
*                               size of the item. Note that this tag must be
*                               used only if the level will be displayed at
*                               the left side or at the right side of the
*                               Slider, otherwise you should pass the
*                               FL_VertLevel tag.
*       FL_VertLevel (ULONG) - If this item will become a Slider with its
*                              level displayed at the gadget's top or bottom
*                              you should specify with this tag the maximum
*                              length of the level string to be displayed.
*                              This is the same value you would pass with
*                              the gadtools GTSL_MaxLevelLen tag. This is
*                              necessary to correctly find out the minimum
*                              size of the item. Note that this tag must be
*                              used only if the level will be displayed at
*                              the top side or at the bottom side of the
*                              Slider, otherwise you should pass the
*                              FL_HorizLevel tag.
*       FL_IndicatorID (ULONG) - If this item will become a ListView with
*                                a string gadget under it, you can specify
*                                with this tag the ID of this string gadget.
*                                Then you can create it with a call to
*                                CreateGadgetLA() just like for the other
*                                gadgets of the layout, even if you didn't
*                                define a real LayoutItem for it.
*                                This is useful because the string gadget
*                                will automatically inherit the width, the
*                                font and the frame height from its "parent"
*                                ListView, so all you have to set before
*                                calling CreateGadgetLA() is its ng_GadgetID.
*
*   RESULT
*       firstchild - pointer to the previous (and possibly current) first
*                    child of this item, or NULL if the item has no children.
*
*   EXAMPLE
*
*       \* Changes the justification of this item *\
*
*       (void)SetLayoutItemAttrs(layout,item,FL_Justify,JT_CENTER,TAG_END);
*
*       \* Changes the sub-tree of children of this item *\
*
*       old = SetLayoutItemAttrs(layout,item,FL_Item,child1,
*                                            FL_Item,child2,TAG_END);
*
*   NOTES
*       If a tag is repeated, only the last instance will be taken into
*       consideration. If two or more conflicting tags are passed, only the
*       last one will be taken into consideration.
*       If you pass a pointer to a LayoutItem which is already part of a
*       linked list, the entire sub-list starting with this item is put into
*       the new sub-tree, while the link with any preceding item is severed.
*
*   CAUTION
*       If you change the sub-tree of a given item, you should always save
*       the returned pointer to the old sub-tree to be able to eventually
*       free it. If you fail to do this, the memory of the old sub-tree won't
*       be deallocated.
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       BuildLItemListA(), GetLayoutItemAttrsA(), CreateLayoutItemA(),
*       SetLayoutAttrsA()
*
****************************************************************************/

struct LayoutItem * ASM SetLayoutItemAttrsA(R_A0 struct Layout *layout, R_A1 struct LayoutItem *li, R_A2 struct TagItem *taglist)
{
   struct LayoutItem *oldfirst = NULL;
   struct TagItem *ti, *tptr;

   if (!li) return (NULL);
   if (!(tptr = taglist)) return (li->First);

   while (ti = NextTagItem(&tptr))
   {
      switch (ti->ti_Tag)
      {
         case FL_AutoSpace:

            if (ti->ti_Data == ~0)
            {
               li->AutoSpacing = 0;
               li->Flags &= ~IF_AUTOSPACE;
            }
            else
            {
               li->AutoSpacing = ti->ti_Data;
               li->Flags |= IF_AUTOSPACE;
            }

            break;

         case FL_Justify:

            li->Flags &= ~IF_JCENTER;

            switch (ti->ti_Data)
            {
               case JT_LEFTTOP:
                  li->Flags |= IF_JLEFTTOP;
                  break;

               case JT_RIGHTBOTTOM:
                  li->Flags |= IF_JRIGHTBOTTOM;
                  break;

               case JT_CENTER:
                  li->Flags |= IF_JCENTER;
            }

            break;

         case FL_AlignItems:

            if (ti->ti_Data)
            {
               li->Flags |= IF_ALIGNITEMS;
            }
            else
            {
               li->Flags &= ~IF_ALIGNITEMS;
            }

            break;

         case FL_HorizLevel:
            li->MaxLevelLen = ti->ti_Data;
            li->LevelPlace = LT_HORIZ;
            break;

         case FL_VertLevel:
            li->MaxLevelLen = ti->ti_Data;
            li->LevelPlace = LT_VERT;
            break;

         case FL_MinSize:
            li->MinSize = ti->ti_Data;
            break;

         case FL_RelSize:
            li->SizeType = ST_RELSPACE;
            li->Size = ti->ti_Data;
            break;

         case FL_Size:
            li->SizeType = ST_FIXED;
            li->Size = ti->ti_Data;
            break;

         case FL_AutoSize:

            switch (li->Type)
            {
               case IT_GADGET:
                  li->SizeType = ST_RELTEXT;
                  break;

               case IT_ZONE:
                  li->SizeType = ST_RELCONT;
                  break;

               case IT_SPACE:
                  li->SizeType = ST_RELSPACE;
            }

            li->Size = (li->Type == IT_SPACE);
            break;

         case FL_BorderType:
            li->BorderType = ti->ti_Data;
            break;

         case FL_BorderWidth:
            li->BorderSizes[0] = ti->ti_Data;
            break;

         case FL_BorderHeight:
            li->BorderSizes[1] = ti->ti_Data;
            break;

         case FL_FrameWidth:
            li->GadgetHDelta = ti->ti_Data + ti->ti_Data;
            break;

         case FL_FrameHeight:
            li->GadgetVDelta = ti->ti_Data + ti->ti_Data;
            break;

         case FL_TextAttr:
            li->TextAttr = (struct TextAttr *)(ti->ti_Data);
            break;

         case FL_GadgetText:
            li->Text = (UBYTE **)(ti->ti_Data);
            break;

         case FL_ListInfo:
            li->ListInfo = (struct ListInfo *)(ti->ti_Data);
            break;

         case FL_Labels:
            li->Labels = (UBYTE **)(ti->ti_Data);
            break;

         case FL_Prev:
            li->Prev = (struct LayoutItem *)(ti->ti_Data);
            break;

         case FL_Next:
            li->Next = (struct LayoutItem *)(ti->ti_Data);
            break;

         case FL_IndicatorID:
            li->IndicatorId = ti->ti_Data;
            break;
      }
   }

   if (li->SizeType == ST_RELTEXT)
   {
      tptr = taglist;

      while (ti = NextTagItem(&tptr))
      {
         switch (ti->ti_Tag)
         {
            case FL_SizeGroup:
               li->GroupId[0] = li->GroupId[1] = ti->ti_Data;
               break;

            case FL_WidthGroup:
               li->GroupId[0] = ti->ti_Data;
               break;

            case FL_HeightGroup:
               li->GroupId[1] = ti->ti_Data;
               break;
         }
      }
   }

   if (li->Type == IT_ZONE)
   {
      oldfirst = li->First;

      if ((ti = FindTagItem(FL_Item,taglist)) && (ti->ti_Data == NULL))
      {
         li->First = NULL;
      }
      else
      {
         if (!(li->First = BuildLItemListA(TRUE,taglist))) li->First = oldfirst;
      }
   }

   return (oldfirst);
}

#ifndef RESLIB

struct LayoutItem *SetLayoutItemAttrs(struct Layout *layout, struct LayoutItem *li, Tag tag1, ...)
{
   return (SetLayoutItemAttrsA(layout,li,(struct TagItem *)&tag1));
}

#endif

/****** fastlayout.library/CreateLayoutItemA ********************************
*
*   NAME
*       CreateLayoutItemA -- Creates a new LayoutItem structure.
*       CreateLayoutItem -- Varargs version of CreateLayoutItemA().
*
*   SYNOPSIS
*       item = CreateLayoutItemA(type, id, layouttype, taglist)
*       D0                       D0    D1  D2          A0
*
*       APTR CreateLayoutItemA(ULONG, ULONG, ULONG, struct TagItem *);
*
*       item = CreateLayoutItem(type, id, layouttype, firsttag, ...)
*
*       APTR CreateLayoutItem(ULONG, ULONG, ULONG, Tag1, ...);
*
*   FUNCTION
*       This function allocates and initializes a new LayoutItem structure.
*       If you pass one or more pointers to LayoutItem structures via the
*       FL_Item tag, these structures will be linked into a list and will
*       become the children of the new item.
*       The varargs version of this function is particularly useful to
*       create an entire LayoutItem tree with a series of nested function
*       calls (see the example below).
*
*   INPUTS
*       type - the type of the new item. The possible types, as of V40, are
*              IT_ZONE, IT_SPACE and IT_GADGET.
*       id - the unique id of the new item. This is important if the item is
*            of type IT_GADGET, as it will allow to associate a gadget with
*            the same id to this item using the CreateGadgetLA() function.
*       layouttype - if this is an IT_ZONE item, this indicates the type
*                    (orientation) of the layout of its sub-items. The
*                    possible layout types are LT_HORIZ and LT_VERT.
*       taglist - pointer to an array of TagItem pairs describing the
*                 attributes of the new LayoutItem structure.
*
*   TAGS
*       FL_AutoSpace (ULONG) - If this tag is present, the sub-items of this
*                              item (which must be a IT_ZONE or the tag will
*                              be ignored) will be automatically spaced.
*                              If there is at least one sub-item of variable
*                              size, the value of this tag is used as the
*                              fixed pixel number to place between sub-items.
*                              If all the sub-items have a fixed size, the
*                              spacing will be variable depending on the size
*                              of the entire layout, and the value of this
*                              tag will be the minimum size of the spacing.
*                              The default is not to do automatic spacing.
*       FL_AlignItems (BOOL) - As of V40, this tag has only meaning for
*                              vertical zones (i.e. items of type IT_ZONE
*                              with a vertical layout), and its purpose is
*                              to align the left (or right) sides of the
*                              gadgets they contain. Currently (V40) this
*                              tag is only guaranteed to work for gadgets
*                              of the kinds BUTTON, STRING, INTEGER, TEXT,
*                              NUMBER, CYCLE, CHECKBOX, SCROLLER and SLIDER
*                              without level indicator.
*                              Not supported yet are the kinds PALETTE,
*                              LISTVIEW and SLIDER with level indicator.
*                              Furthermore, for this tag to work, the
*                              justification of the gadgets must be JT_FLUSH.
*                              The default is FALSE (no alignment).
*       FL_Item (APTR) - Pointer to a valid LayoutItem structure, which will
*                        become a child of this LayoutItem. You can pass more
*                        than one FL_Item tag. It is incorrect to pass a
*                        FL_Item tag if this is not a IT_ZONE.
*       FL_GadgetText (STRPTR *) - Pointer to the pointer to the string this
*                                  item, when actualized into a gadget, will
*                                  contain. The double indirection is useful
*                                  to be able to change the string pointer
*                                  without having to reassign it to the
*                                  LayoutItem. It is incorrect to pass a
*                                  FL_GadgetText if this is not a IT_GADGET.
*                                  If you pass a NULL here, or not use this
*                                  tag, there should be no text in the
*                                  gadget, or its minimum size will be
*                                  calculated incorrectly.
*                                  The default is NULL.
*       FL_TextAttr (struct TextAttr *) - Pointer to a valid TextAttr
*                                         structure describing the font that
*                                         will be used for the gadget
*                                         associated with this LayoutItem.
*                                         If you pass a NULL here, or not
*                                         use this tag, the font of
*                                         the Layout structure will be
*                                         used instead, so if all your
*                                         gadgets use the same font, you
*                                         don't have to specify this font
*                                         for each and every LayoutItem in
*                                         the tree. It is incorrect to pass
*                                         a FL_TextAttr tag if this is not a
*                                         IT_GADGET.
*                                         The default is NULL.
*       FL_FrameWidth (UWORD) - Width of the frame of the gadget associated
*                               to this item. This is the width of ONE of the
*                               two (left and right) vertical borders of the
*                               gadget. Note that this tag is only valid for
*                               AutoSize gadgets. If you pass -1 (or ~0), or
*                               not use this tag, this item will use the
*                               default value specified for the Layout
*                               structure. It is incorrect to pass a
*                               FL_FrameWidth tag if this is not a IT_GADGET.
*                               The default is to use the value of the Layout
*                               structure.
*       FL_FrameHeight (UWORD) - Height of the frame of the gadget associated
*                                to this item. This is the height of ONE of
*                                the two (top and bottom) horizontal borders
*                                of the gadget. Note that this tag is only
*                                valid for AutoSize gadgets. If you pass -1
*                                (or ~0), or not use this tag, this item will
*                                use the default value specified for the
*                                Layout structure. It is incorrect to pass
*                                a FL_FrameHeight tag if this is not an
*                                IT_GADGET.
*                                The default is to use the value of the
*                                Layout structure.
*       FL_BorderType (ULONG) - Type of border this item must have. This is
*                               used mainly for IT_ZONE items, as the gadgets
*                               already have a border. The layout will take
*                               the size of the border into account; to
*                               actually render the border you must call the
*                               RefreshLayout() function. The currently
*                               defined types are:
*
*                               BT_NONE - No border.
*                               BT_SPACE - Invisible border.
*                               BT_RAISED - The standard gadtools border.
*                               BT_RECESSED - As above, but recessed.
*                               BT_LINE - A furrow (recessed gadtools ridge).
*
*                               Note that there no BT_RIDGE type yet. This
*                               could change in the future if gadtools will
*                               support a multi-line string-gadget without
*                               borders.
*                               The default is BT_NONE.
*       FL_BorderWidth (UWORD) - Width of the border of this item, if it has
*                                one. This refers to the border added by the
*                                library with FL_BorderType. If you pass -1
*                                (or ~0), or not use this tag, the value of
*                                the Layout structure will be used.
*                                The default is to use the value of the
*                                Layout structure.
*       FL_BorderHeight (UWORD) - Height of the border of this item, if it
*                                 has one. This refers to the border added by
*                                 the library with FL_BorderType. If you pass
*                                 -1 (or ~0), or not use this tag, the value
*                                 of the Layout structure will be used.
*                                 The default is to use the value of the
*                                 Layout structure.
*       FL_Justify (ULONG) - Type of justification of this item in the
*                            secundary direction, that is the direction
*                            perpendicular to the one of the layout of the
*                            parent item. The types currently defined are:
*
*                            JT_FLUSH - Spread the item using all the
*                                       available space.
*                            JT_LEFTTOP - Place the item to the left (top)
*                                         side of its parent item.
*                            JT_RIGHTBOTTOM - Place the item to the right
*                                             (bottom) side of its parent
*                                             item.
*                            JT_CENTER - Place the item in the center of
*                                        its parent item.
*
*                            With the exception of the JT_FLUSH case, the
*                            item will be reduced to its minimum secundary
*                            size when justified.
*                            The default is JT_FLUSH.
*       FL_Size (ULONG) - Fixed absolute size in pixels of this item in the
*                         direction of the layout of its parent item. If
*                         you specify this tag, the item will always have
*                         the same size, i. e. won't adapt to the font (if
*                         a IT_GADGET) or to its contents (if a IT_ZONE).
*                         The default is to make the item adapt to contents.
*       FL_RelSize (ULONG) - Size of this item in the direction of its
*                            parent's layout, in relation to the space
*                            available (the size of the parent item minus
*                            the size of all fixed-size items). This is the
*                            "weight" of this item, so the actual size will
*                            depend on the relative size of the other items
*                            with the FL_RelSize tag. For example, if you
*                            want only two items, one of which has always a
*                            size equal to the half of the other one's, you
*                            can specify a FL_RelSize of 2 and 1 for the
*                            items; you could also specify 100 and 50.
*                            The actual size is also limited by the minimum
*                            size of the item, so when the layout is at its
*                            minimum size the FL_RelSize gadgets won't
*                            respect their relative weights (unless the
*                            minimum sizes, too, reflect them).
*                            The default is to make the item adapt to
*                            contents (size not variable).
*       FL_AutoSize (ULONG) - If you specify this tag the item will assume
*                             a size (in the direction of its parent's
*                             layout) that is the minimum to allow its
*                             contents to fit into it. In other words, the
*                             IT_GADGET items will be sized taking into
*                             account their text, font, frame and border
*                             size, and other specific attributes; the
*                             IT_ZONE items will be sized taking into account
*                             the sub-items they contain. The IT_SPACE items
*                             will always be set to their minimum size as
*                             specified with the FL_MinSize tag.
*                             Note that to determine the size in the
*                             secundary direction you must use the FL_Justify
*                             tag and pass the appropriate tags to the
*                             parent(s) of this item.
*                             As of V40 the actual value passed with this tag
*                             is ignored, so you should pass a value of zero.
*                             This tag is optional, that is, if you do not
*                             pass any of the FL_Size, FL_RelSize and
*                             FL_AutoSize tags, FL_AutoSize will be assumed.
*       FL_MinSize (ULONG) - The absolute minimum size this item can reach.
*                            It is guaranteed that the item will never be
*                            smaller than this, no matter which other tags
*                            are used. This refers only to the direction of
*                            the layout of this item's parent; to determine
*                            the minimum size in the secundary direction you
*                            must set it in the parents themselves.
*                            The default is 10.
*       FL_Labels (STRPTR *) - If this item will become a list-type gadget,
*                              that is (currently) a ListView, a MX or a
*                              Cycle gadget, you should pass with this tag
*                              the pointer to the array of labels the gadget
*                              will display. This is necessary to correctly
*                              find out the minimum size of the item.
*                              The array must be NULL-terminated.
*                              Note: as of V40, this tag is not required for
*                              items which will become ListViews.
*                              The default is NULL.
*       FL_ListInfo (struct ListInfo *) - If this item will become a
*                                         list-type gadget, that is
*                                         (currently) a ListView, a MX or a
*                                         Cycle gadget, you should pass with
*                                         this tag the pointer to a ListInfo
*                                         structure describing the attributes
*                                         of the gadget. The structure fields
*                                         (and their meaning) are:
*
*                                         GadgetKind - The exact kind of
*                                                      the gadget. It can
*                                                      be LISTVIEW_KIND,
*                                                      MX_KIND or CYCLE_KIND.
*
*                                         TextPlace - For MX gadgets this
*                                                     indicates where the
*                                                     labels must go with
*                                                     respect to the buttons.
*                                                     For the other two kinds
*                                                     this indicates the
*                                                     position of the gadget
*                                                     text. The values can be
*                                                     the usual PLACETEXT_#?
*                                                     definitions.
*
*                                         Spacing - The amount of spacing
*                                                   between labels of a
*                                                   MX or ListView gadget.
*                                                   This is ignored for
*                                                   Cycle gadgets.
*
*                                         GadgetWidth - For MX gadgets, this
*                                                       is the width of the
*                                                       buttons in pixels.
*                                                       If your program runs
*                                                       under an OS previous
*                                                       to 3.0, this value
*                                                       will be ignored and
*                                                       MX_WIDTH (17) will be
*                                                       assumed. If you put 0
*                                                       here, under 3.0+ an
*                                                       automatic value will
*                                                       be used (namely font
*                                                       height + 1).
*                                                       For ListView gadgets,
*                                                       this is the width of
*                                                       the scrollbar; if you
*                                                       put 0 here, you will
*                                                       get the default (16).
*                                                       This field is ignored
*                                                       for Cycle gadgets.
*
*                                         GadgetHeight - For MX gadgets, this
*                                                        is the height of the
*                                                        buttons in pixels.
*                                                        If your program runs
*                                                        under an OS previous
*                                                        to 3.0, this value
*                                                        will be ignored and
*                                                        MX_HEIGHT (9) will
*                                                        be assumed. If you
*                                                        put 0 here, under
*                                                        3.0+ an automatic
*                                                        value will be used
*                                                        (namely font height
*                                                        + 1).
*                                                        For ListView gadgets
*                                                        this is the height
*                                                        of the indicator
*                                                        placed under the
*                                                        gadget (if there is
*                                                        one). If you put
*                                                        0 here, it will be
*                                                        assumed an height
*                                                        equal to the font
*                                                        height plus the
*                                                        double of the value
*                                                        passed with the
*                                                        FL_FrameHeight tag.
*                                                        It is recommended
*                                                        that you always use
*                                                        this value when you
*                                                        create the string
*                                                        gadget to place
*                                                        under your ListView.
*                                                        This field is
*                                                        ignored for Cycle
*                                                        gadgets.
*
*                                         Flags - Currently (V40) only one
*                                                 flag is defined:
*                                                 LIF_ADJUSTTOP causes a
*                                                 ListView to be vertically
*                                                 centered inside its item.
*                                                 This is useful because
*                                                 GadTools may shorten the
*                                                 ListView to make it contain
*                                                 an integral number of
*                                                 labels, and this way the
*                                                 extra space gets equally
*                                                 distributed on both sides
*                                                 of the ListView.
*                                                 This flag is ignored for
*                                                 MX and Cycle gadgets.
*                                         The default is NULL.
*       FL_WidthGroup (ULONG) - If you want two or more items to have the
*                               same width, you can specify this tag for each
*                               of them, passing the same (arbitrary) number.
*                               This will form an ideal "class" of items
*                               having the same width.
*                               Note: currently this works only for items
*                               of type IT_GADGET with the FL_AutoSize tag.
*                               The default is 0 (no width class).
*       FL_HeightGroup (ULONG) - If you want two or more items to have the
*                                same height, you can specify this tag for
*                                each of them, passing the same (arbitrary)
*                                number.
*                                This will form an ideal "class" of items
*                                having the same height.
*                                Note: currently this works only for items
*                                of type IT_GADGET with the FL_AutoSize tag.
*                                The default is 0 (no height class).
*       FL_SizeGroup (ULONG) - If you want two or more items to have the
*                              same size, you can specify this tag for each
*                              of them, passing the same (arbitrary) number.
*                              This will form an ideal "class" of items
*                              having the same size. This is equivalent to
*                              passing both the FL_WidthGroup and
*                              FL_HeightGroup tags with the same number.
*                              Note: currently this works only for items
*                              of type IT_GADGET with the FL_AutoSize tag.
*                              The default is 0 (no size class).
*       FL_HorizLevel (ULONG) - If this item will become a Slider with its
*                               level displayed at the gadget's left or right
*                               you should specify with this tag the maximum
*                               length of the level string to be displayed.
*                               This is the same value you would pass with
*                               the gadtools GTSL_MaxLevelLen tag. This is
*                               necessary to correctly find out the minimum
*                               size of the item. Note that this tag must be
*                               used only if the level will be displayed at
*                               the left side or at the right side of the
*                               Slider, otherwise you should pass the
*                               FL_VertLevel tag.
*                               The default is no level displayed.
*       FL_VertLevel (ULONG) - If this item will become a Slider with its
*                              level displayed at the gadget's top or bottom
*                              you should specify with this tag the maximum
*                              length of the level string to be displayed.
*                              This is the same value you would pass with
*                              the gadtools GTSL_MaxLevelLen tag. This is
*                              necessary to correctly find out the minimum
*                              size of the item. Note that this tag must be
*                              used only if the level will be displayed at
*                              the top side or at the bottom side of the
*                              Slider, otherwise you should pass the
*                              FL_HorizLevel tag.
*                              The default is no level displayed.
*       FL_IndicatorID (ULONG) - If this item will become a ListView with
*                                a string gadget under it, you can specify
*                                with this tag the ID of this string gadget.
*                                Then you can create it with a call to
*                                CreateGadgetLA() just like for the other
*                                gadgets of the layout, even if you didn't
*                                define a real LayoutItem for it.
*                                This is useful because the string gadget
*                                will automatically inherit the width, the
*                                font and the frame height from its "parent"
*                                ListView, so all you have to set before
*                                calling CreateGadgetLA() is its ng_GadgetID.
*                                The default is obviously 0, meaning this
*                                LayoutItem will have no indicator.
*
*   RESULT
*       item - pointer to the new LayoutItem structure, or NULL if the new
*              item couldn't be created.
*
*   EXAMPLE
*
*       \* Creates a layout consisting of three gadgets, as follows: *\
*       \*                                                           *\
*       \*               +---------------------------+               *\
*       \*               |                           |               *\
*       \*               |    Item with variable     |               *\
*       \*               |    height, suitable       |               *\
*       \*               |    for a ListView         |               *\
*       \*               |                           |               *\
*       \*               |                           |               *\
*       \*               +---------------------------+               *\
*       \*               | Button |  Auto   | Button |               *\
*       \*               |   #1   | Spacing |   #2   |               *\
*       \*               +---------------------------+               *\
*
*       ...
*
*       #define LISTVIEW_ID 1
*       #define BUTT1_ID    2
*       #define BUTT2_ID    3
*
*       #define NewItem CreateLayoutItem
*
*       root = NewItem(IT_ZONE,0,LT_VERT,
*                      FL_RelSize,1,
*                      FL_Item,NewItem(IT_GADGET,LISTVIEW_ID,0,
*                                      FL_RelSize,1,
*                                      FL_ListInfo,&listinfo,
*                                      TAG_END),
*                      FL_Item,NewItem(IT_ZONE,0,LT_HORIZ,
*                                      FL_AutoSpace,4,
*                                      FL_Item,NewItem(IT_GADGET,BUTT1_ID,0,
*                                                      FL_TextAttr,&txat,
*                                                      FL_Text,&text1,
*                                                      FL_WidthGroup,1,
*                                                      TAG_END),
*                                      FL_Item,NewItem(IT_GADGET,BUTT2_ID,0,
*                                                      FL_TextAttr,&txat,
*                                                      FL_Text,&text2,
*                                                      FL_WidthGroup,1,
*                                                      TAG_END),
*                                      TAG_END),
*                      TAG_END);
*
*       ...
*
*   NOTES
*       In the include file "libraries/fastlayout.h" is defined a shortcut
*       for CreateLayoutItemA() (and CreateLayoutItem()), as seen above.
*       It is suggested you use this shortcut, NewItemA() (and NewItem()),
*       to make your source code more readable and, why not, to save some
*       typing...
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       GetLayoutItemAttrsA(), SetLayoutItemAttrs(), FreeLayoutItems()
*
****************************************************************************/

ULONG ASM CreateLayoutItemA(R_D0 ULONG type, R_D1 ULONG id, R_D2 ULONG ltype, R_A0 struct TagItem *taglist)
{
   struct LayoutItem *li;

   if (!(li = AllocVec(sizeof(struct LayoutItem),MEMF_CLEAR)))
   {
      return ((ULONG)NULL);
   }

   switch (type)
   {
      case IT_GADGET:
         li->SizeType = ST_RELTEXT;
         break;

      case IT_ZONE:
         li->SizeType = ST_RELCONT;
         break;

      case IT_SPACE:
         li->SizeType = ST_RELSPACE;
   }

   li->Size = (type == IT_SPACE);
   li->MinSize = (type == IT_GADGET)? DEFMINSIZE: 0;
   li->Type = type;
   li->LayoutType = ltype;
   li->GadgetHDelta = 0xFFFE;
   li->GadgetVDelta = 0xFFFE;
   li->BorderSizes[0] = li->BorderSizes[1] = -1;
   li->Id = id;

   (void)SetLayoutItemAttrsA(NULL,li,taglist);

   return ((ULONG)li);
}

#ifndef RESLIB

ULONG CreateLayoutItem(ULONG type, ULONG id, ULONG ltype, Tag tag1, ...)
{
   return (CreateLayoutItemA(type,id,ltype,(struct TagItem *)&tag1));
}

#endif

#ifdef RESLIB
static ULONG MaxLabelWidth(struct TextAttr *ta, UBYTE **labels)
#else
static ULONG __asm MaxLabelWidth(register __a0 struct TextAttr *ta, register __a1 UBYTE **labels)
#endif
{
   ULONG width, maxwidth = 0L;

   // Trova la larghezza massima dei labels.

   while (*labels)
   {
      if ((width = TextWidth(ta,*(labels++))) > maxwidth) maxwidth = width;
   }

   return (maxwidth);
}

static ULONG __regargs MinListSize(struct LayoutItem *item, ULONG orientation, struct Layout *layout)
{
   ULONG minsize = 0L, numlabels = 0L, textwidth = 0L;
   UWORD gadghdelta, gadgvdelta, mxwidth, mxheight, ysize;
   struct ListInfo *li;
   struct TextAttr *ta;
   struct Library *SysBase = *((struct Library **)4L);

   // Ricava alcuni parametri: font del gadget, altezza di tale font,
   // larghezza del testo del gadget (se c') e spessore dei bordi.

   li = item->ListInfo;
   GetTextAttr(&ta,item,layout);
   ysize = ta->ta_YSize;
   if (item->Text) textwidth = TextWidth(ta,*(item->Text));
   if ((gadghdelta = item->GadgetHDelta) & 0x8000) gadghdelta = (layout? layout->GadgetHDelta: DEFGADGHDELTA);
   if ((gadgvdelta = item->GadgetVDelta) & 0x8000) gadgvdelta = (layout? layout->GadgetVDelta: DEFGADGVDELTA);

   // Se  stato fornito l'array di label, conta quanti sono.
   // Questo serve solo ai gadget di tipo MX, e in questo caso l'array
   // di label esiste sicuramente (il controllo  fatto dalla funzione
   // chiamante).

   if (li->GadgetKind == MX_KIND)
   {
      for (; item->Labels[numlabels]; numlabels++);
   }

   // Calcola la dimensione minima nella direzione richiesta in base al
   // tipo di gadget e alle caratteristiche dell'eventuale testo del gadget.

   if (orientation == LT_HORIZ)
   {
      switch (li->GadgetKind)
      {
         case MX_KIND:
            minsize = MaxLabelWidth(ta,item->Labels);
            mxwidth = ((SysBase->lib_Version < 39)? MX_WIDTH: li->GadgetWidth);
            if (!mxwidth) mxwidth = ysize + 1;
            if (mxwidth < MX_WIDTH) mxwidth = MX_WIDTH;

            if (!(li->TextPlace) || (li->TextPlace == PLACETEXT_LEFT))
            {
               minsize += mxwidth + 8;
            }
            else if (li->TextPlace == PLACETEXT_RIGHT)
            {
               minsize += mxwidth + 7;
            }
            else if (mxwidth > minsize)
            {
               minsize = mxwidth;
            }

            break;

         case LISTVIEW_KIND:
            minsize = (li->GadgetWidth? li->GadgetWidth: 16) + 16;

            if (item->Text && *(item->Text))
            {
               if (li->TextPlace == PLACETEXT_LEFT)
               {
                  minsize += textwidth + 8;
               }
               else if (li->TextPlace == PLACETEXT_RIGHT)
               {
                  minsize += textwidth + 7;
               }
               else if (textwidth > minsize)
               {
                  minsize = textwidth;
               }
            }

            break;

         case CYCLE_KIND:
            minsize = MaxLabelWidth(ta,item->Labels) + gadghdelta + 20;

            if (item->Text && *(item->Text))
            {
               if (!(li->TextPlace) || (li->TextPlace == PLACETEXT_LEFT))
               {
                  minsize += textwidth + 8;
               }
               else if (li->TextPlace == PLACETEXT_RIGHT)
               {
                  minsize += textwidth + 7;
               }
               else if (textwidth > minsize)
               {
                  minsize = textwidth;
               }
            }

            break;
      }
   }
   else
   {
      switch (li->GadgetKind)
      {
         case MX_KIND:
            minsize = (ysize + li->Spacing) * numlabels - li->Spacing;
            mxheight = ((SysBase->lib_Version < 39)? MX_HEIGHT: li->GadgetHeight);
            if (!mxheight) mxheight = ysize + 1;
            if (mxheight < MX_HEIGHT) mxheight = MX_HEIGHT;

            if (li->TextPlace == PLACETEXT_ABOVE)
            {
               minsize += mxheight + 8;
            }
            else if (li->TextPlace == PLACETEXT_BELOW)
            {
               minsize += mxheight + 7;
            }
            else if (mxheight > ysize)
            {
               minsize += mxheight - ysize;
            }

            break;

         case LISTVIEW_KIND:
            minsize = ysize + li->Spacing + 4;
            minsize += (li->GadgetHeight? li->GadgetHeight: ysize + gadgvdelta);

            if (item->Text && *(item->Text))
            {
               if (!(li->TextPlace) || (li->TextPlace == PLACETEXT_ABOVE))
               {
                  minsize += ysize + 8;
               }
               else if (li->TextPlace == PLACETEXT_BELOW)
               {
                  minsize += ysize + 7;
               }
               else if (ysize > minsize)
               {
                  minsize = ysize;
               }
            }

            break;

         case CYCLE_KIND:
            minsize = ysize + gadgvdelta;

            if (item->Text && *(item->Text))
            {
               if (li->TextPlace == PLACETEXT_ABOVE)
               {
                  minsize += ysize + 8;
               }
               else if (li->TextPlace == PLACETEXT_BELOW)
               {
                  minsize += ysize + 7;
               }
               else if (ysize > minsize)
               {
                  minsize = ysize;
               }
            }

            break;
      }
   }

   return (minsize);
}

#ifdef RESLIB
static ULONG MinGadgetSize(struct LayoutItem *item, ULONG orientation, ULONG parentltype, struct Layout *layout)
#else
static ULONG __asm MinGadgetSize(register __a0 struct LayoutItem *item, register __d0 ULONG orientation, register __d1 ULONG parentltype, register __a1 struct Layout *layout)
#endif
{
   ULONG minsizetot;
   LONG minsize;
   UWORD gadghdelta, gadgvdelta;
   UBYTE *text = NULL;
   struct TextAttr *ta;

   // Se il gadget  di tipo LISTVIEW, MX o CYCLE ne calcola la dimensione
   // minima in base ai parametri supplementari forniti nella struttura
   // ListInfo, altrimenti tiene conto solo del testo e del font specificati.
   // Nota: se il gadget  di tipo MX o CYCLE  indispensabile conoscere
   // l'array di labels; se non c' si tratta il gadget come se fosse di
   // tipo normale, con risultati meno corretti.

   if (item->ListInfo && (item->Labels || (item->ListInfo->GadgetKind == LISTVIEW_KIND)))
   {
      minsizetot = MinListSize(item,orientation,layout);
   }
   else
   {
      // Trova lo spessore dei bordi del gadget. I valori di default,
      // naturalmente, sono DEFGADGHDELTA e DEFGADGVDELTA.

      if ((gadghdelta = item->GadgetHDelta) & 0x8000) gadghdelta = (layout? layout->GadgetHDelta: DEFGADGHDELTA);
      if ((gadgvdelta = item->GadgetVDelta) & 0x8000) gadgvdelta = (layout? layout->GadgetVDelta: DEFGADGVDELTA);

      // Poich il testo del gadget  uno solo, nel caso si trovi all'esterno
      // il gadget stesso pu assumere dimensioni piccole a piacere, o quasi.

      GetTextAttr(&ta,item,layout);
      if (item->Text) text = *(item->Text);
      minsizetot = (orientation? (ta->ta_YSize + gadgvdelta): (TextWidth(ta,text) + gadghdelta));

      // Se si tratta di uno slider aut similia, tiene conto anche dello
      // spazio occupato dall'eventuale stringa LevelFormat.

      if (item->MaxLevelLen && (item->LevelPlace == orientation))
      {
         minsizetot += 8 + (orientation? ta->ta_YSize: (item->MaxLevelLen * FontWidth(ta)));
      }
   }

   // Tiene conto delle dimensioni dell'eventuale bordo assegnato al gadget

   if (item->BorderType)
   {
      if ((minsize = item->BorderSizes[orientation]) < 0) minsize = (layout? layout->BorderSizes[orientation]: (orientation? DEFBVSIZE: DEFBHSIZE));
      minsizetot += minsize + minsize;
   }

   // Tiene conto anche dell'eventuale dimensione minima assegnata al gadget
   // dal programmatore (default DEFMINSIZE).

   minsize = (parentltype == orientation)? item->MinSize: 0;
   if (minsize > minsizetot) minsizetot = minsize;

   return (minsizetot);
}

#ifdef RESLIB
static ULONG MinGroupSize(struct LayoutItem *item, ULONG gid, ULONG orientation, ULONG parentltype, struct Layout *layout)
#else
static ULONG __asm MinGroupSize(register __a0 struct LayoutItem *item, register __d0 ULONG gid, register __d1 ULONG orientation, register __d2 ULONG parentltype, register __a1 struct Layout *layout)
#endif
{
   struct LayoutItem *li;
   ULONG minsize, minsizetot = 0L;

   // Per ogni item, se  un gadget ST_RELTEXT che appartiene allo stesso
   // gruppo a larghezza/altezza uguale, calcola la dimensione minima e
   // tiene traccia della massima tra queste; se invece  una zona con dei
   // sotto-item, applica questa funzione ricorsivamente su di essi,
   // ricavando la dimensione minima massima degli eventuali altri item di
   // questo tipo, di cui tiene conto.
   // Se l'item  invece uno spazio, una zona senza sotto-item o un gadget
   // di altro tipo, non fa niente.

   for (li = item; li; li = li->Next)
   {
      if (li->GroupId[orientation] == gid)
      {
         minsize = MinGadgetSize(li,orientation,parentltype,layout);
      }
      else if (li->First)
      {
         minsize = MinGroupSize(li->First,gid,orientation,parentltype,layout);
      }
      else
      {
         minsize = 0;
      }

      if (minsize > minsizetot) minsizetot = minsize;
   }

   return (minsizetot);
}

#ifdef RESLIB
static ULONG MinSize(struct LayoutItem *item, ULONG orientation, ULONG parentltype, struct Layout *layout, struct LayoutItem *root)
#else
static ULONG __asm MinSize(register __a0 struct LayoutItem *item, register __d0 ULONG orientation, register __d1 ULONG parentltype, register __a1 struct Layout *layout, register __a2 struct LayoutItem *root)
#endif
{
   ULONG minsize, minsizetot = 0L;
   UWORD count = 0;
   LONG bsize = 0;
   struct LayoutItem *li;
   struct TextAttr *ta;

   if (!item) return (0L);

   // Se l'item  ST_FIXED (a dimensione fissa specificata in pixel) si
   // prende la sua dimensione come quella minima, altrimenti si calcola
   // la dimensione minima in base al tipo di item e ad altri parametri.

   if ((item->SizeType == ST_FIXED) && (parentltype == orientation))
   {
      // In questo caso la dimensione minima coincide con la dimensione
      // fissa (essendo l'item ST_FIXED).

      minsizetot = item->Size;
   }
   else
   {
      // In base al tipo dell'item (che non  ST_FIXED) calcola
      // la sua dimensione minima nella direzione specificata.

      if (item->Type == IT_ZONE)
      {
         // Trova la dimensione minima di tutti gli item contenuti in
         // questa zona e ne fa la somma se gli item sono disposti secondo
         // la direzione specificata, altrimenti prende la dimensione
         // minima massima.

         for (li = item->First; li; li = li->Next)
         {
            minsize = MinSize(li,orientation,item->LayoutType,layout,root);

            if (item->LayoutType == orientation)
            {
               minsizetot += minsize;
            }
            else
            {
               if (minsize > minsizetot) minsizetot = minsize;
            }

            count += 1;
         }

         // Se per questa zona  stata richiesta la spaziatura automatica,
         // ne tiene conto nel determinare la dimensione minima.

         if ((item->LayoutType == orientation) && (item->Flags & IF_AUTOSPACE))
         {
            minsizetot += (item->AutoSpacing * (count - 1));
         }

         // Inoltre tiene conto anche della dimensione dell'eventuale bordo
         // richiesto per questa zona.

         if (item->BorderType)
         {
            if ((bsize = item->BorderSizes[orientation]) < 0) bsize = (layout? layout->BorderSizes[orientation]: (orientation? DEFBVSIZE: DEFBHSIZE));
            minsizetot += bsize + bsize;
         }

         // Tiene anche conto dell'eventuale titolo di questa zona

         if (item->Text)
         {
            GetTextAttr(&ta,item,layout);

            if (orientation == LT_HORIZ)
            {
               minsize = bsize + bsize + TextWidth(ta,*(item->Text));
               if (minsize > minsizetot) minsizetot = minsize;
            }
            else
            {
               minsizetot += ta->ta_YSize;
            }
         }

         // Infine tiene conto della dimensione minima assegnata dal
         // programmatore (default 0).

         minsize = (parentltype == orientation)? item->MinSize: 0;
         if (minsize > minsizetot) minsizetot = minsize;
      }
      else if (item->Type == IT_GADGET)
      {
         // Se non  un item ST_RELTEXT che deve avere la stessa dimensione
         // di altri dello stesso tipo, calcola la sua dimensione minima
         // nel modo normale. Altrimenti calcola la dimensione minima
         // di ogni item dello stesso tipo e di queste prende la massima.
         // Questo vale per entrambe le direzioni, indipendentemente dalla
         // direzione del layout dell'item che fa da contenitore.
         // Notare che se l'item non  ST_RELTEXT, GroupId[0|1]  sempre 0.

         if (item->GroupId[orientation] == 0)
         {
            // Trova la dimensione minima di questo item (gadget)

            minsizetot = MinGadgetSize(item,orientation,parentltype,layout);
         }
         else
         {
            // Trova la massima dimensione minima degli item ST_RELTEXT
            // che hanno lo stesso GroupId di questo item.

            minsizetot = MinGroupSize(root,item->GroupId[orientation],orientation,parentltype,layout);
         }
      }
      else if (item->Type == IT_SPACE)
      {
         // L'unica dimensione minima che pu avere uno spazio  quella
         // assegnata dal programmatore (default 0).

         minsizetot = ((parentltype == orientation)? item->MinSize: 0L);
      }
   }

   return (minsizetot);
}

/****** fastlayout.library/MinimumWidth *************************************
*
*   NAME
*       MinimumWidth -- Computes the minimum width of a layout.
*
*   SYNOPSIS
*       minwidth = MinimumWidth(layout, root)
*       D0                      A0      A1
*
*       ULONG MinimumWidth(APTR, APTR);
*
*   FUNCTION
*       The purpose of this function is to allow your program to know the
*       minimum horizontal space your layout needs to be rendered properly.
*       For this function to work correctly, you need to specify for each
*       item of the layout all the revelant size-related information, such
*       as, for instance, the TextAttr and GadgetText for IT_GADGET items.
*       The main use of this function is to find out what must be the
*       minimum width of the window which will contain your layout.
*       If your window is not resizable, you could also use the result of
*       this function to assign the window its exact width.
*
*   INPUTS
*       layout - pointer to a context reference Layout structure.
*       root - pointer to the root of your LayoutItem tree.
*
*   RESULT
*       minwidth - the minimum width (in pixels) of your layout.
*
*   EXAMPLE
*
*       ...
*       minwidth = MinimumWidth(lt,root) + scr->WBorLeft + scr->WBorRight;
*       win = OpenWindowTags(NULL,WA_Screen,scr,WA_MinWidth,minwidth, ...);
*       ...
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       MinimumHeight(), CreateLayoutItemA(), SetLayoutItemAttrsA()
*
****************************************************************************/

ULONG ASM MinimumWidth(R_A0 struct Layout *layout, R_A1 struct LayoutItem *root)
{
   return (MinSize(root,LT_HORIZ,(layout? layout->LayoutType: LT_HORIZ),layout,root));
}

/****** fastlayout.library/MinimumHeight ************************************
*
*   NAME
*       MinimumHeight -- Computes the minimum height of a layout.
*
*   SYNOPSIS
*       minheight = MinimumHeight(layout, root)
*       D0                        A0      A1
*
*       ULONG MinimumHeight(APTR, APTR);
*
*   FUNCTION
*       The purpose of this function is to allow your program to know the
*       minimum vertical space your layout needs to be rendered properly.
*       For this function to work correctly, you need to specify for each
*       item of the layout all the revelant size-related information, such
*       as, for instance, the TextAttr and GadgetText for IT_GADGET items.
*       The main use of this function is to find out what must be the
*       minimum height of the window which will contain your layout.
*       If your window is not resizable, you could also use the result of
*       this function to assign the window its exact height.
*
*   INPUTS
*       layout - pointer to a context reference Layout structure.
*       root - pointer to the root of your LayoutItem tree.
*
*   RESULT
*       minheight - the minimum height (in pixels) of your layout.
*
*   EXAMPLE
*
*       ...
*       winbarheight = scr->WBorTop + scr->Font->ta_YSize + 1;
*       minheight = MinimumHeight(lt,root) + winbarheight + scr->WBorBottom;
*       win = OpenWindowTags(NULL,WA_Screen,scr,WA_MinHeight,minheight, ...);
*       ...
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       MinimumWidth(), CreateLayoutItemA(), SetLayoutItemAttrsA()
*
****************************************************************************/

ULONG ASM MinimumHeight(R_A0 struct Layout *layout, R_A1 struct LayoutItem *root)
{
   return (MinSize(root,LT_VERT,(layout? layout->LayoutType: LT_HORIZ),layout,root));
}

static ULONG LayoutItems(struct LayoutItem *item, ULONG left, ULONG top, ULONG width, ULONG height, ULONG ltype, ULONG flags, ULONG autospacing, struct Layout *layout, struct LayoutItem *root)
{
   struct LayoutItem *li, *li2;
   struct TextAttr *ta;
   LONG delta, hbsize, vbsize;
   ULONG fixedspace = 0L, totalrelsize = 0L;
   ULONG zonesize, zonespace, usedspace;
   ULONG size, minsize, cleft, ctop, cwidth, cheight, jtype;
   ULONG ok = FALSE;
   WORD shift;
   UWORD count = 0, spacetype = 0;

   if (!item || !width || !height) return (FALSE);

   // Calcola lo spazio fisso occupato e trova numero e tipo degli elementi

   for (li = item; li; li = li->Next)
   {
      count += 1;

      if (li->SizeType == ST_FIXED)
      {
         fixedspace += li->Size;
      }
      else if ((li->SizeType == ST_RELTEXT) || (li->SizeType == ST_RELCONT))
      {
         fixedspace += MinSize(li,ltype,ltype,layout,root);
      }
      else  // ST_RELSPACE
      {
         spacetype = 1;
         totalrelsize += li->Size;
      }
   }

   if ((flags & IF_AUTOSPACE) && spacetype) fixedspace += autospacing * (count - 1);

   // Calcola lo spazio libero a disposizione degli elementi

   zonesize = ((ltype == LT_HORIZ)? width: height);  // Spazio totale
   if (fixedspace > zonesize) return (FALSE);
   zonespace = zonesize - fixedspace;          // Spazio per item variabili

   // Calcola la posizione e le dimensioni reali degli elementi

   while (!ok)
   {
      usedspace = 0L;
      ok = TRUE;               // Siamo ottimisti...

      for (li = item; li; li = li->Next)
      {
         minsize = MinSize(li,ltype,ltype,layout,root);
         GetTextAttr(&ta,li,layout);

         // Se l'item NON  di tipo ST_RELSPACE, oppure lo  ma ha il
         // flag IF_MINSIZE (per cui DEVE usare la sua dimensione minima)
         // allora si assume come sua dimensione corrente quella minima.
         // Questo vale anche per gli item di tipo ST_FIXED in quanto la
         // loro dimensione (fissa) coincide con quella minima.

         if ((li->SizeType != ST_RELSPACE) || (li->Flags & IF_MINSIZE))
         {
            size = minsize;  // minsize == li->Size quando  ST_FIXED
         }
         else  // Questo item ha dim. variabile
         {
            size = (zonespace * li->Size) / totalrelsize;

            // Se la dimensione assegnata  inferiore alla minima consentita,
            // si impone all'item di usare la minima consentita e lo si
            // considera come a dimensione fissa. Si sottrae quindi la sua
            // dimensione allo spazio disponibile per gli item variabili e
            // il suo peso relativo viene sottratto da totalrelsize. Poi
            // si ricomincia da capo il ciclo per adattare gli altri item.
            // Tutto questo pu accadere solo per gli item con ST_RELSPACE.

            if (size < minsize)
            {
               li->Flags |= IF_MINSIZE;
               zonespace -= minsize;
               totalrelsize -= li->Size;
               ok = FALSE;
               break;
            }
         }

         // Qui si calcolano la posizione e le dimensioni effettive,
         // tenendo conto dell'eventuale giustificazione richiesta,
         // con un piccolo aggiustamento se questo  l'ultimo elemento e
         // il bordo destro/inferiore non coincide perfettamente con
         // quello della zona contenitore.
         // Nota: questo dovrebbe essere fatto SOLO se esistono almeno due
         // item a dimensione variabile, altrimenti non avrebbe senso poich:
         //  Item tutti a dimensione fissa => dimensione contenitore esatta;
         //  Un solo item a dimensione variabile => sua dimensione esatta.
         // In questo caso si considerano item a dimensione variabile anche
         // gli spazi automatici.

         jtype = li->Flags & IF_JCENTER;

         if (ltype == LT_HORIZ)
         {
            cleft = left + usedspace;
            cwidth = size;

            cheight = ((jtype == JT_FLUSH)? height: MinSize(li,1 - ltype,1 - ltype,layout,root));
            ctop = top;

            if (jtype == JT_RIGHTBOTTOM)
            {
               ctop += (height - cheight);
            }
            else  if (jtype == JT_CENTER)
            {
               ctop += (height - cheight) / 2;
            }

            if ((!li->Next) && (spacetype || (flags & IF_AUTOSPACE)))
            {
               delta = (left + width) - (cleft + cwidth);

               while (delta > 0)
               {
                  shift = 0;

                  for (li2 = item; li2->Next; li2 = li2->Next)
                  {
                     li2->Left += shift;

                     if (delta)
                     {
                        if (li2->SizeType == ST_RELSPACE)
                        {
                           li2->Width += 1;
                        }

                        if ((li2->SizeType == ST_RELSPACE) || !spacetype)
                        {
                           shift += 1;
                           delta -= 1;
                        }
                     }
                  }

                  cleft += shift;
               }
            }
         }
         else
         {
            ctop = top + usedspace;
            cheight = size;

            cwidth = ((jtype == JT_FLUSH)? width: MinSize(li,1 - ltype,1 - ltype,layout,root));
            cleft = left;

            if (jtype == JT_RIGHTBOTTOM)
            {
               cleft += (width - cwidth);
            }
            else if (jtype == JT_CENTER)
            {
               cleft += (width - cwidth) / 2;
            }

            if ((!li->Next) && (spacetype || (flags & IF_AUTOSPACE)))
            {
               delta = (top + height) - (ctop + cheight);

               while (delta > 0)
               {
                  shift = 0;

                  for (li2 = item; li2->Next; li2 = li2->Next)
                  {
                     li2->Top += shift;

                     if (delta)
                     {
                        if (li2->SizeType == ST_RELSPACE)
                        {
                           li2->Height += 1;
                        }

                        if ((li2->SizeType == ST_RELSPACE) || !spacetype)
                        {
                           shift += 1;
                           delta -= 1;
                        }
                     }
                  }

                  ctop += shift;
               }
            }
         }

         usedspace += size;

         // Se  stata richiesta la spaziatura automatica, si tiene conto
         // della sua dimensione nel disporre gli item.

         if ((flags & IF_AUTOSPACE) && li->Next)
         {
            usedspace += (spacetype? autospacing: (zonespace / (count - 1)));
         }

         // Posizione e dimensioni vengono modificate per tenere conto
         // dell'eventuale bordo richiesto per questo item.

         if (li->BorderType)
         {
            if ((hbsize = li->BorderSizes[LT_HORIZ]) < 0) hbsize = (layout? layout->BorderSizes[LT_HORIZ]: DEFBHSIZE);
            if ((vbsize = li->BorderSizes[LT_VERT]) < 0) vbsize = (layout? layout->BorderSizes[LT_VERT]: DEFBVSIZE);

            cleft += hbsize;
            ctop += vbsize;
            cwidth -= (hbsize + hbsize);
            cheight -= (vbsize + vbsize);
         }

         if ((li->Type == IT_ZONE) && (li->Text))
         {
            ctop += ta->ta_YSize;
            cheight -= ta->ta_YSize;
         }

         // Salviamo nell'item i suoi dati correnti.

         li->Left = cleft;
         li->Top = ctop;
         li->Width = cwidth;
         li->Height = cheight;
      }
   }

   // Rimuoviamo dagli item il flag IF_MINSIZE che ora non serve pi.
   // Inoltre per ogni item che ha dei sottoitem effettuiamo il layout
   // anche di questi.

   ok = TRUE;

   for (li = item; li; li = li->Next)
   {
      li->Flags &= ~IF_MINSIZE;

      if ((li->Type == IT_ZONE) && li->First)
      {
         if (!(LayoutItems(li->First,li->Left,li->Top,li->Width,li->Height,li->LayoutType,li->Flags,li->AutoSpacing,layout,root))) ok = FALSE;
      }
   }

   return (ok);
}

/****** fastlayout.library/ComputeLayout ************************************
*
*   NAME
*       ComputeLayout -- Computes position and size of each LayoutItem.
*
*   SYNOPSIS
*       success = ComputeLayout(layout, root, left, top, width, height)
*       D0                      A0      A1    D0    D1   D2     D3
*
*       ULONG ComputeLayout(APTR, APTR, ULONG, ULONG, ULONG, ULONG);
*
*   FUNCTION
*       This function computes the actual layout of the items in your tree
*       in terms of the position and size of the rectangle which must
*       contain it. Each LayoutItem in the tree will get an actual position
*       and size in pixels, and the layout will be ready to be used to
*       create the associated gadgets with the CreateGadgetLA() function.
*       To use this function you must first create a tree of LayoutItem
*       structures with the CreateLayoutItemA() function, and possibly a
*       valid Layout structure with the CreateLayoutA() function.
*       The Layout structure can be omitted, in which case the global
*       library defaults will be used for any undefined LayoutItem
*       attribute.
*
*   INPUTS
*       layout - pointer to a reference Layout structure, or NULL.
*       root - pointer to the root of your LayoutItem tree.
*       left - left edge of the layout rectangle, relative to its window.
*       top - top edge of the layout rectangle, relative to its window.
*       width - width of the layout rectangle in pixels.
*       height - height of the layout rectangle in pixels.
*
*   RESULT
*       success - TRUE if all went OK, FALSE if there were errors. The most
*                 common cause of errors is the presence in the LayoutItem
*                 tree of a zone too small to contain its assigned sub-items.
*
*   EXAMPLE
*
*       ...
*       left = mywin->BorderLeft;
*       top = mywin->BorderTop;
*       width = mywin->Width - mywin->BorderLeft - mywin->BorderRight;
*       height = mywin->Height - mywin->BorderTop - mywin->BorderBottom;
*
*       ok = ComputeLayout(mylayout,myroot,left,top,width,height);
*
*       if (!ok) printf("Error in the layout!\n");
*       ...
*
*   NOTES
*       You MUST call this function at least once before you call
*       CreateGadgetLA() with your layout, or unpredictable things could
*       happen, from invisible gadgets to a system crash.
*       Furthermore, note that you can specify the initial direction of
*       your layout by passing the appropriate value (LT_HORIZ or LT_VERT)
*       to CreateLayoutA(). So if you don't use a Layout structure you
*       will get the default, that is LT_HORIZ.
*       It is recommended that you use a Layout structure, as if you don't
*       the default font will be Topaz 8, which is less than desirable.
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       CreateLayoutA(), CreateLayoutItemA(), CreateGadgetLA()
*
****************************************************************************/

ULONG ASM ComputeLayout(R_A0 struct Layout *layout, R_A1 struct LayoutItem *root, R_D0 ULONG left, R_D1 ULONG top, R_D2 ULONG width, R_D3 ULONG height)
{
   return (LayoutItems(root,left,top,width,height,(layout? layout->LayoutType: LT_HORIZ),0L,0L,layout,root));
}

/****** fastlayout.library/CreateGadgetLA ***********************************
*
*   NAME
*       CreateGadgetLA -- Replaces CreateGadgetA() and follows a layout.
*       CreateGadgetL -- Varargs version of CreateGadgetLA().
*
*   SYNOPSIS
*       gad = CreateGadgetLA(layout, root, kind, previous, newgad, tagList)
*       D0                   A3      A4    D0    A0        A1      A2
*
*       struct Gadget *CreateGadgetLA(APTR, APTR, ULONG, struct Gadget *,
*                                     struct NewGadget *, struct TagItem *);
*
*       gad = CreateGadgetL(layout, root, kind, previous, newgad, tag1, ...)
*
*       struct Gadget *CreateGadgetL(APTR, APTR, ULONG, struct Gadget *,
*                                    struct NewGadget *, Tag, ...);
*
*   FUNCTION
*       This function is meant to replace the analogous CreateGadgetA()
*       function of the gadtools.library to take advantage of the information
*       contained in a LayoutItem tree.
*       The use of the function is identical to that of the gadtools
*       function, except for what follows: you supply to CreateGadgetLA() a
*       pointer to the root of a valid LayoutItem tree which describes your
*       layout, and possibly also a pointer to a reference Layout structure.
*       In the NewGadget structure whose pointer you pass to the function you
*       don't have to specify the ng_Left, ng_Top, ng_Width and ng_Height
*       values: CreateGadgetLA() automatically positions and sizes the new
*       gadget accordingly to the size and position of the item of the tree
*       which is relative to this gadget. For this to be possible, the
*       ng_GadgetID of the new gadget must match that of the corresponding
*       LayoutItem.
*       If a LayoutItem of type IT_GADGET with an Id equal to ng_GadgetID is
*       found, a new gadget is created with the position and size of the
*       item. If no such LayoutItem is found, no gadget is created.
*       This means that if you want to create gadgets independently from the
*       layout, you must use the traditional CreateGadgetA() function.
*       The ng_TextAttr field, too, can be set to NULL, which means that the
*       TextAttr of the corresponding LayoutItem will be automatically used
*       for the new gadget.
*       Apart from ng_Left, ng_Top, ng_Width, ng_Height and ng_TextAttr, all
*       of the other fields of the NewGadget structure, as well as all of
*       the tags passed to the function, will be preserved, recognized and
*       used just as always.
*       The first four of the aforementioned fields, instead, after the call
*       will contain the actual position and size of the gadget just created.
*       For further information on the creation of the gadgets, refer to
*       the autodoc of the gadtools.library/CreateGadgetA() function.
*
*   IMPLEMENTATION
*       The compatibility with gadtools is guaranteed by the fact that the
*       CreateGadgetLA() function does indeed call CreateGadgetA() after
*       doing its own internal processing.
*
*   INPUTS
*       layout - pointer to a reference Layout structure, or NULL if you
*                just want to use the defaults for undefined attributes.
*       root - pointer to the root of your LayoutItem tree.
*       kind - kind of gadget to be created, one of the XXXX_KIND values
*              defined in <libraries/gadtools.h>.
*       previous - pointer to the previous gadget that this new gadget
*                  is to be attached to. This function will fail if this
*                  value is NULL.
*       newgad - a filled in NewGadget structure describing the desired
*                gadget's attributes, except for size, position and font.
*       tagList - pointer to an array of tags providing optional extra
*                 parameters, or NULL.
*
*   TAGS
*       For the list of tags to be used with this function, refer to the
*       autodoc of the gadtools.library/CreateGadgetA() function.
*
*   RESULT
*       gad - pointer to the new gadget, or NULL if the allocation failed
*             or if previous was NULL or if there wasn't a LayoutItem with
*             a matching Id in the tree.
*
*   EXAMPLE
*
*       gad = CreateGadgetL(layout,root,BUTTON_KIND,gad,&newgadget,TAG_END);
*
*   NOTES
*       There is a significant exception to what said above: in the case of
*       MX (radio button) gadgets, after the call to CreateGadgetLA() the
*       ng_Width and ng_Height will contain the actual width and height of
*       the single buttons, and not that of the entire gadget. To obtain the
*       latter, you can use GetLayoutItemAttrsA() on the LayoutItem whose
*       Id matches the MX gadget's one (you find this LayoutItem's pointer
*       using the FindLayoutItem() function as usual).
*       The values you will find in the ng_Width and ng_Height fields will be
*       the same you put into the ListInfo structure whose pointer you passed
*       to the corresponding LayoutItem with the FL_ListInfo tag. However, if
*       you specify a value of 0 in the GadgetWidth and GadgetHeight fields
*       of the ListInfo structure, or if there is no ListInfo structure, the
*       function will automatically calculate an appropriate value for the
*       buttons' width and height (currently this value is the font height
*       plus one pixel).
*       However, if your program runs under a version of AmigaOS previous
*       to V39, or if you don't specify the { GTMX_Scaled, TRUE } tag item,
*       the defaults (MX_WIDTH and MX_HEIGHT, namely 17 and 9) will be used
*       instead of your values.
*       Note that the ng_VisualInfo field of the NewGadget structure must be
*       set to a valid VisualInfo pointer, or this function will fail.
*       The ng_TextAttr field can instead be NULL, and in this case the
*       TextAttr of the corresponding LayoutItem will be used.
*       If the TextAttr is not NULL, it should be the same passed to the
*       corresponding LayoutItem via the FL_TextAttr tag.
*
*   CAUTION
*       You MUST call ComputeLayout() on your LayoutItem tree at least
*       once before calling this function, or unpredictable things could
*       happen, from invisible gadgets to a system crash.
*
*   BUGS
*       None known. Obviously this function has the same bugs of the
*       gadtools.library/CreateGadgetA() function.
*
*   SEE ALSO
*       ComputeLayout(), gadtools.library/CreateGadgetA(),
*       gadtools.library/FreeGadgets(), gadtools.library/GT_SetGadgetAttrs(),
*       gadtools.library/GetVisualInfo(), <libraries/gadtools.h>
*
****************************************************************************/

struct Gadget * ASM CreateGadgetLA(R_A3 struct Layout *layout, R_A4 struct LayoutItem *root, R_D0 ULONG kind, R_A0 struct Gadget *gad, R_A1 struct NewGadget *newgadget, R_A2 struct TagItem *taglist)
{
   struct LayoutItem *li, *tmp;
   struct TagItem *ti;
   struct NewGadget ngcopy, *ng = &ngcopy;
   struct Library *SysBase = *((struct Library **)4L);
   struct TextAttr *ta;
   ULONG textplace, size, asize = 0;
   UWORD gadgetheight, fontheight, labelheight, ssheight = 0;
   UWORD gadgvdelta, levelplace, adjusted = 0;
   BOOL align = FALSE;

   if (!root || !gad || !newgadget || !(newgadget->ng_VisualInfo)) return (NULL);

   ngcopy = *newgadget;

   // Trova la struttura LayoutItem corrispondente a questo gadget

   if (!(li = FindLayoutItem(root,IT_GADGET,ng->ng_GadgetID)))
   {
      if (!((li = FindIndicator(root,ng->ng_GadgetID)) && ((kind == STRING_KIND) || (kind == INTEGER_KIND))))
      {
         return (NULL);
      }
   }

   // Controlla se si debbano allineare i gadget in verticale

   if (tmp = FindLayoutItemParent(NULL,root,IT_GADGET,ng->ng_GadgetID))
   {
      align = ((tmp->Flags & IF_ALIGNITEMS) && (tmp->LayoutType == LT_VERT));
   }

   // Qui si calcolano la posizione e le dimensioni del nuovo gadget
   // in base alle informazioni contenute nella struttura LayoutItem
   // corrispondente, nella struttura NewGadget ed eventualmente anche
   // nei TagItem.

   ng->ng_LeftEdge = li->Left;
   ng->ng_TopEdge = li->Top;
   if (kind != MX_KIND) ng->ng_Width = li->Width;
   if (kind != MX_KIND) ng->ng_Height = li->Height;

   if (!ng->ng_TextAttr)
   {
      GetTextAttr(&(ng->ng_TextAttr),li,layout);
   }

   if (align)
   {
      for (tmp = li; tmp->Prev; tmp = tmp->Prev) ;

      for (; tmp; tmp = tmp->Next)
      {
         if (tmp->Text)
         {
            GetTextAttr(&ta,tmp,layout);
            if ((size = TextWidth(ta,*(tmp->Text))) > asize) asize = size;
         }
      }
   }

   if ((textplace = ng->ng_Flags & 0x0000001F) == 0L)
   {
      textplace = ((kind == LISTVIEW_KIND) || (kind == PALETTE_KIND))? PLACETEXT_ABOVE: PLACETEXT_LEFT;
   }

   if ((gadgvdelta = li->GadgetVDelta) & 0x8000) gadgvdelta = (layout? layout->GadgetVDelta: DEFGADGVDELTA);

   if (ng->ng_GadgetText && (kind != MX_KIND))
   {
      if (textplace & PLACETEXT_ABOVE)
      {
         ng->ng_TopEdge += (ng->ng_TextAttr->ta_YSize + 8);
         ng->ng_Height -= (ng->ng_TextAttr->ta_YSize + 8);
         adjusted = PLACETEXT_ABOVE;
      }
      else if (textplace & PLACETEXT_BELOW)
      {
         ng->ng_Height -= (ng->ng_TextAttr->ta_YSize + 7);
         adjusted = PLACETEXT_BELOW;
      }
      else if (textplace & PLACETEXT_LEFT)
      {
         size = (align? asize: TextWidth(ng->ng_TextAttr,ng->ng_GadgetText));
         ng->ng_LeftEdge += (size + 8);
         ng->ng_Width -= (size + 8);
         adjusted = PLACETEXT_LEFT;
      }
      else if (textplace & PLACETEXT_RIGHT)
      {
         size = (align? asize: TextWidth(ng->ng_TextAttr,ng->ng_GadgetText));
         ng->ng_Width -= (size + 7);
         adjusted = PLACETEXT_RIGHT;
      }
   }

   if ((kind == SLIDER_KIND) && li->MaxLevelLen)
   {
      if ((levelplace = GetTagData(GTSL_LevelPlace,PLACETEXT_LEFT,taglist)) != adjusted)
      {
         if (levelplace == PLACETEXT_ABOVE)
         {
            ng->ng_TopEdge += (ng->ng_TextAttr->ta_YSize + 8);
            ng->ng_Height -= (ng->ng_TextAttr->ta_YSize + 8);
         }
         else if (levelplace == PLACETEXT_BELOW)
         {
            ng->ng_Height -= (ng->ng_TextAttr->ta_YSize + 7);
         }
         else if (levelplace == PLACETEXT_LEFT)
         {
            size = li->MaxLevelLen * FontWidth(ng->ng_TextAttr);
            ng->ng_LeftEdge += (size + 8);
            ng->ng_Width -= (size + 8);
         }
         else if (levelplace == PLACETEXT_RIGHT)
         {
            size = li->MaxLevelLen * FontWidth(ng->ng_TextAttr);
            ng->ng_Width -= (size + 7);
         }
      }
   }

   if (kind == LISTVIEW_KIND)
   {
      if (li->ListInfo && (li->ListInfo->Flags & LIF_ADJUSTTOP))
      {
         fontheight = ng->ng_TextAttr->ta_YSize;
         labelheight = fontheight + GetTagData(LAYOUTA_Spacing,0,taglist);

         if (ti = FindTagItem(GTLV_ShowSelected,taglist))
         {
            if (ti->ti_Data)
            {
               ssheight = ((struct Gadget *)(ti->ti_Data))->Height + gadgvdelta;
            }
            else if (SysBase->lib_Version < 39)
            {
               ssheight = (fontheight + gadgvdelta);
            }
         }

         gadgetheight = ng->ng_Height - (ssheight + 4);
         gadgetheight = (gadgetheight / labelheight) * labelheight + 4 + ssheight;

         ng->ng_TopEdge += ((ng->ng_Height - gadgetheight) / 2);
         ng->ng_Height = gadgetheight;
      }
   }

   if (kind == MX_KIND)
   {
      if ((SysBase->lib_Version < 39) || !(GetTagData(GTMX_Scaled,FALSE,taglist)))
      {
         ng->ng_Width = MX_WIDTH;
         ng->ng_Height = MX_HEIGHT;
      }
      else
      {
         ng->ng_Width = (li->ListInfo? li->ListInfo->GadgetWidth: 0);
         if (ng->ng_Width <= 0) ng->ng_Width = ng->ng_TextAttr->ta_YSize + 1;
         if (ng->ng_Width < MX_WIDTH) ng->ng_Width = MX_WIDTH;

         ng->ng_Height = (li->ListInfo? li->ListInfo->GadgetHeight: 0);
         if (ng->ng_Height <= 0) ng->ng_Height = ng->ng_TextAttr->ta_YSize + 1;
         if (ng->ng_Height < MX_HEIGHT) ng->ng_Height = MX_HEIGHT;
      }
   }

   if (kind == CHECKBOX_KIND)
   {
      if ((SysBase->lib_Version < 39) || !(GetTagData(GTCB_Scaled,FALSE,taglist)))
      {
         if (textplace & (PLACETEXT_RIGHT | PLACETEXT_LEFT))
         {
            ng->ng_TopEdge += 1;
         }
      }
   }

   /* Se si tratta di uno string gadget che diventer */
   /* l'indicatore di un gadget di tipo ListView      */

   if (((kind == STRING_KIND) || (kind == INTEGER_KIND)) && (li->Id != ng->ng_GadgetID))
   {
      ng->ng_LeftEdge = ng->ng_TopEdge = 0;
      ng->ng_Width = li->Width;
      ng->ng_Height = ng->ng_TextAttr->ta_YSize + gadgvdelta;
      ng->ng_GadgetText = NULL;
   }

   if (ng->ng_Width < 0) ng->ng_Width = 0;
   if (ng->ng_Height < 0) ng->ng_Height = 0;

   *((struct IBox *)newgadget) = *((struct IBox *)ng);

   //newgadget->ng_LeftEdge = ng->ng_LeftEdge;
   //newgadget->ng_TopEdge = ng->ng_TopEdge;
   //newgadget->ng_Width = ng->ng_Width;
   //newgadget->ng_Height = ng->ng_Height;

   return (CreateGadgetA(kind,gad,ng,taglist));
}

#ifndef RESLIB

struct Gadget *CreateGadgetL(struct Layout *layout, struct LayoutItem *root, ULONG kind, struct Gadget *gad, struct NewGadget *ng, Tag tag1, ...)
{
   return (CreateGadgetLA(layout,root,kind,gad,ng,(struct TagItem *)&tag1));
}

#endif

