

/*  GetLayoutAttrsA.c  $ 31/05/95 MT $  */


#include "exec/types.h"
#include "graphics/text.h"
#include "intuition/intuition.h"
#include "utility/tagitem.h"
#include "proto/utility.h"
#include "rl_fastlayout.h"
#include "fastlayout.h"
#include "fl_private.h"


extern struct TextAttr *FL_Private1;


/****** fastlayout.library/GetLayoutAttrsA **********************************
*
*   NAME
*       GetLayoutAttrsA -- Inquires about the attributes of a Layout.
*       GetLayoutAttrs -- Varargs version of GetLayoutAttrsA().
*
*   SYNOPSIS
*       GetLayoutAttrsA(layout, taglist)
*                       A0      A1
*
*       void GetLayoutAttrsA(APTR, struct TagItem *);
*
*       GetLayoutAttrs(layout, firsttag, ...)
*
*       void GetLayoutAttrs(APTR, Tag, ...);
*
*   FUNCTION
*       This function allows you to inquire about the attributes of a Layout
*       structure. The values of most of these attributes are used for the
*       corresponding attributes of any LayoutItem which does not have them
*       already defined.
*       Not all of the attributes can be read with this function; the
*       attributes you can inquire about are mainly those you also can
*       change with SetLayoutAttrsA().
*       You ask the value of a particular attribute by passing its tag
*       whose tagvalue is a pointer to a variable of the appropriate type.
*       After the function call, you can read that variable to find out the
*       value of the required attribute.
*
*   INPUTS
*       layout - pointer to the Layout structure whose attributes are to
*                be inquired about.
*       taglist - pointer to an array of TagItem pairs.
*
*   TAGS
*       FL_TextAttr (struct TextAttr **) - Pointer to the TextAttr structure
*                                          that will be used as default for
*                                          the items of this layout.
*       FL_FrameWidth (ULONG *) - Width of the frame of gadgets in this
*                                 layout which do not have the FL_FrameWidth
*                                 tag. This is the width of ONE of the two
*                                 (left and right) vertical borders of the
*                                 gadgets.
*       FL_FrameHeight (ULONG *) - Height of the frame of gadgets in this
*                                  layout which don't have the FL_FrameHeight
*                                  tag. This is the height of ONE of the two
*                                  (top and bottom) horizontal borders of the
*                                  gadgets.
*       FL_BorderWidth (ULONG *) - Width of the border of items with a border
*                                  but without the FL_BorderWidth tag.
*       FL_BorderHeight (ULONG *) - Height of the border of items with a
*                                   border but without the FL_BorderHeight
*                                   tag.
*
*   EXAMPLE
*
*       \* Finds out the TextAttr assigned to this reference Layout *\
*
*       struct TextAttr *ta;
*
*       GetLayoutAttrs(layout,FL_TextAttr,&ta,TAG_END);
*
*   NOTES
*       If a tag is repeated, only the last instance will be taken into
*       consideration.
*       If you pass a NULL Layout pointer, the function will return the
*       global library default for the attributes you inquire about.
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       SetLayoutAttrsA(), CreateLayoutA(), GetLayoutItemAttrsA()
*
****************************************************************************/

void ASM GetLayoutAttrsA(R_A0 struct Layout *layout, R_A1 struct TagItem *taglist)
{
   struct TagItem *ti;
#ifdef RESLIB
   struct TagItem *tptr = taglist;
#endif
   ULONG *value;

#ifdef RESLIB
   while (ti = NextTagItem(&tptr))
#else
   while (ti = NextTagItem(&taglist))
#endif
   {
      if (!(value = (ULONG *)(ti->ti_Data))) continue;

      switch (ti->ti_Tag)
      {
         case FL_TextAttr:
            *value = (ULONG)(layout? layout->TextAttr: FL_Private1);
            break;

         case FL_FrameWidth:
            *value = (ULONG)((layout? layout->GadgetHDelta: DEFGADGHDELTA) / 2);
            break;

         case FL_FrameHeight:
            *value = (ULONG)((layout? layout->GadgetVDelta: DEFGADGVDELTA) / 2);
            break;

         case FL_BorderWidth:
            *value = (ULONG)(layout? layout->BorderSizes[LT_HORIZ]: DEFBHSIZE);
            break;

         case FL_BorderHeight:
            *value = (ULONG)(layout? layout->BorderSizes[LT_VERT]: DEFBVSIZE);
            break;
      }
   }
}

#ifndef RESLIB

void GetLayoutAttrs(struct Layout *layout, Tag tag1, ...)
{
   GetLayoutAttrsA(layout,(struct TagItem *)&tag1);
}

#endif

