

/*  RefreshLayout.c  $ 31/05/95 MT $  */


#include "exec/types.h"
#include "exec/libraries.h"
#include "intuition/intuition.h"
#include "proto/gadtools.h"
#include "proto/intuition.h"
#include "rl_fastlayout.h"
#include "fastlayout.h"
#include "fl_private.h"


extern struct TextAttr *FL_Private1;
extern ULONG (*FL_Private2)(struct TextAttr *, UBYTE *);


#ifdef RESLIB
static void RefreshLayoutItems(struct Window *win, APTR vi, struct LayoutItem *item, struct Layout *layout)
#else
static void __regargs RefreshLayoutItems(struct Window *win, APTR vi, struct LayoutItem *item, struct Layout *layout)
#endif
{
   struct LayoutItem *li;
   struct TextAttr *ta;
   struct IntuiText itext = { 0 };
   LONG hbsize, vbsize, dtop;

   for (li = item; li; li = li->Next)
   {
      dtop = 0;

      if ((li->Type == IT_ZONE) && li->Text)
      {
         if (!(ta = li->TextAttr)) ta = (layout? layout->TextAttr: FL_Private1);
         dtop = ta->ta_YSize / 2 + 1;
      }

      if (li->BorderType)
      {
         if ((hbsize = li->BorderSizes[LT_HORIZ]) < 0) hbsize = (layout? layout->BorderSizes[LT_HORIZ]: DEFBHSIZE);
         if ((vbsize = li->BorderSizes[LT_VERT]) < 0) vbsize = (layout? layout->BorderSizes[LT_VERT]: DEFBVSIZE);

         switch (li->BorderType)
         {
            case BT_RAISED:
            case BT_RECESSED:
               DrawBevelBox(win->RPort,li->Left - hbsize,li->Top - vbsize - dtop,li->Width + hbsize + hbsize,li->Height + vbsize + vbsize + dtop,GT_VisualInfo,vi,(li->BorderType == BT_RECESSED)? GTBB_Recessed: TAG_IGNORE,TRUE,TAG_END);
               break;

            case BT_LINE:
               DrawBevelBox(win->RPort,li->Left - hbsize,li->Top - vbsize - dtop,li->Width + hbsize + hbsize,li->Height + vbsize + vbsize + dtop,GT_VisualInfo,vi,GTBB_Recessed,TRUE,GTBB_FrameType,BBFT_RIDGE,TAG_END);
               if (GadToolsBase->lib_Version < 39) DrawBevelBox(win->RPort,li->Left - hbsize + 2,li->Top - vbsize - dtop + 1,li->Width + hbsize + hbsize - 4,li->Height + vbsize + vbsize + dtop - 2,GT_VisualInfo,vi,TAG_END);
         }
      }

      if (dtop)
      {
         itext.FrontPen = 2;  /* Dovrebbe realmente usare un DrawInfo! */
         itext.DrawMode = JAM2;
         itext.LeftEdge = li->Left + ((ULONG)(li->Width - FL_Private2(ta,*(li->Text)))) >> 1;
         itext.TopEdge = li->Top - vbsize - ta->ta_YSize;
         itext.ITextFont = ta;
         itext.IText = *(li->Text);

         PrintIText(win->RPort,&itext,0,0);
      }

      RefreshLayoutItems(win,vi,li->First,layout);
   }
}

/****** fastlayout.library/RefreshLayout ************************************
*
*   NAME
*       RefreshLayout -- Renders the layout imagery.
*
*   SYNOPSIS
*       RefreshLayout(window, layout, root)
*                     A0      A1      A2
*
*       void RefreshLayout(struct Window *, APTR, APTR);
*
*   FUNCTION
*       This function draws into your window all of the borders you put
*       into your layout with the FL_BorderType tag.
*       You should call this function when doing the other normal refresh
*       operations, as well as when your gadget list is first displayed.
*
*   INPUTS
*       window - pointer to your window. It can be NULL, in which case
*                nothing is done.
*       layout - pointer to your Layout structure. It can be NULL, in which
*                case the defaults are used for undefined attributes.
*       root - pointer to the root of your LayoutItem tree. It can be NULL,
*              in which case nothing is done.
*
*   CAUTION
*       You MUST call ComputeLayout() on your LayoutItem tree at least
*       once before calling this function, or unpredictable things could
*       happen, from invisible borders to a system crash.
*
*   BUGS
*       None known.
*
*   SEE ALSO
*       CreateLayoutItemA()
*
****************************************************************************/

void ASM RefreshLayout(R_A0 struct Window *win, R_A1 struct Layout *layout, R_A2 struct LayoutItem *root)
{
   APTR vi;

   if (!win || !root) return;

   if (vi = GetVisualInfoA(win->WScreen,NULL))
   {
      RefreshLayoutItems(win,vi,root,layout);
      FreeVisualInfo(vi);
   }
}

