

#ifndef LIBRARIES_FASTLAYOUT_H
#define LIBRARIES_FASTLAYOUT_H


/*  Fast Layout Library  $ 15/05/95 MT $  */
/*  Header file for standard definitions  */


#ifndef EXEC_TYPES_H
#include "exec/types.h"
#endif

#ifndef GRAPHICS_RASTPORT_H
#include "graphics/rastport.h"
#endif


/* Short names for CreateLayoutItemA() and CreateLayoutItem() */

#define NewItemA CreateLayoutItemA
#define NewItem  CreateLayoutItem

/* The tags recognized by the Fast Layout functions */

#define FL_TagBase      TAG_USER
#define FL_AutoSpace    (FL_TagBase + 1)  /* Automatically space the items */
#define FL_Item         (FL_TagBase + 2)  /* Pointer to LayoutItem */
#define FL_GadgetText   (FL_TagBase + 3)  /* Text this item will contain */
#define FL_TextAttr     (FL_TagBase + 4)  /* TextAttr of the gadget text */
#define FL_FrameWidth   (FL_TagBase + 5)  /* Gadget frame width */
#define FL_FrameHeight  (FL_TagBase + 6)  /* Gadget frame height */
#define FL_BorderType   (FL_TagBase + 7)  /* Type of border of this item */
#define FL_Justify      (FL_TagBase + 8)  /* Type of justification */
#define FL_Size         (FL_TagBase + 9)  /* Absolute size in pixels */
#define FL_RelSize      (FL_TagBase + 10) /* Relative size */
#define FL_AutoSize     (FL_TagBase + 11) /* Automatically size to fit */
#define FL_MinSize      (FL_TagBase + 12) /* Minimum size of this item */
#define FL_Labels       (FL_TagBase + 13) /* Labels for a cycle or mx */
#define FL_ListInfo     (FL_TagBase + 14) /* Extra info for a list item */
#define FL_WidthGroup   (FL_TagBase + 15) /* ID of a "same width" group */
#define FL_HeightGroup  (FL_TagBase + 16) /* ID of a "same height" group */
#define FL_SizeGroup    (FL_TagBase + 17) /* ID of a "same size" group */
#define FL_HorizLevel   (FL_TagBase + 18) /* Max horiz level string length */
#define FL_VertLevel    (FL_TagBase + 19) /* Max vert level string length */
#define FL_ItemSpec     (FL_TagBase + 20) /* For enquiring about pos/size */
#define FL_AlignItems   (FL_TagBase + 21) /* Align the gadgets vertically */
#define FL_BorderWidth  (FL_TagBase + 22) /* Border width of the item */
#define FL_BorderHeight (FL_TagBase + 23) /* Border height of the item */
#define FL_Prev         (FL_TagBase + 24) /* Previous item in the list */
#define FL_Next         (FL_TagBase + 25) /* Next item in the list */
#define FL_IndicatorID  (FL_TagBase + 26) /* ListView's string gadget ID */
#define FL_BorderTitle  (FL_GadgetText)   /* Not yet implemented */

/* Layout types */

#define LT_HORIZ 0      /* Layout the items horizontally */
#define LT_VERT  1      /* Layout the items vertically   */

/* Border types */

#define BT_NONE     0   /* No border */
#define BT_SPACE    1   /* Border is a simple offset */
#define BT_RAISED   2   /* Border is a raised gadtools bevelled box */
#define BT_RECESSED 3   /* Border is a recessed gadtools bevelled box */
#define BT_LINE     4   /* Border is a furrow (recessed gadtools ridge) */

/* Justification types */

#define JT_FLUSH       0  /* Flush the item (default) */
#define JT_LEFTTOP     1  /* Item must stay left- or top-aligned */
#define JT_RIGHTBOTTOM 2  /* Item must stay right- or bottom-aligned */
#define JT_CENTER      3  /* Item must be centered */

/* LayoutItem types */

#define IT_ZONE   0     /* This item is a zone */
#define IT_SPACE  1     /* This item is a space */
#define IT_GADGET 2     /* This item is a gadget */

/* LayoutItem tree scanning types */

#define SCAN_NODEFIRST 0  /* Implement a sort of pre-order tree scanning */
#define SCAN_NODELAST  1  /* Implement a sort of post-order tree scanning */

/* Label list types */

#define LL_STRING  0
#define LL_INTEGER 1

/* ListInfo flags */

#define LIF_ADJUSTTOP 0x00000001  /* Center listview in LayoutItem */


/* ListInfo structure for list-type gadgets: */
/* use for MX, CYCLE and LISTVIEW kind.      */

struct ListInfo
{
   UWORD GadgetKind;    /* Exact kind of the gadget */
   ULONG TextPlace;     /* Same values as NewGadget.ng_Flags */
   UWORD Spacing;       /* Spacing between items in pixels */
   UWORD GadgetWidth;   /* Width of buttons (MX) or scrollbar (LISTVIEW) */
   UWORD GadgetHeight;  /* Height of buttons (MX) or indicator (LISTVIEW) */
   ULONG Flags;         /* General-purpose flags (see values above) */
};


#endif

