
#ifndef CLIB_SAVAGE_H
#define CLIB_SAVAGE_H

#include <exec/types.h>
#include <exec/lists.h>
#include <exec/nodes.h>
#include <exec/resident.h>
#include <exec/ports.h>
#include <exec/tasks.h>
#include <intuition/intuition.h>
#include <intuition/screens.h>
#include <dos/dos.h>
#include <graphics/text.h>
#include <graphics/gfx.h>
#include <savage/sysinfo.h>
#include <savage/guicontext.h>
#include <savage/visible.h>
#include <utility/tagitem.h>

/********************************************************************/

// datei 
APTR sav_ReadFile(char *);
APTR sav_ReadFile2(char *);
BOOL sav_FileExists(char *);
BOOL sav_DirExists(char *);
struct FileInfoBlock *sav_FileGetFIB(char *);
void sav_FileFreeFIB(struct FileInfoBlock *);
long sav_FileSize(char *);
long sav_FileType(char *);
long sav_FileSum(char *);
char *sav_FileVersion(char *);
BOOL sav_FileWriteProtected(char *);
long sav_FileUnProtect(char *);
struct Resident *sav_GetFileRomtag(char *);
void sav_FreeFileRomtag(void);
void sav_Print(char *);
void sav_PrintLF(void);
BPTR sav_FindCOS(void);
BPTR sav_FindCIS(void);
BPTR sav_CD(char *);
long sav_PathLen(char *);
BPTR sav_MakeDir(char *);
BOOL sav_FontFile(char *);
BOOL sav_InfoFile(char *);
BOOL sav_DirWalk(char *, BOOL (*)(struct FileInfoBlock *, APTR), APTR);
struct FileLock *sav_GetActualDir(void);
void sav_ReduceDir(char *);
long sav_GetProtection(char *);

// gui
struct GUIContext *sav_GetGUIContext(char *);
void sav_FreeGUIContext(struct GUIContext *);
void sav_WindowInnerDim(struct Window *, struct Rectangle *);
void sav_CloseSharedWindow(struct Window *);
void sav_VisibleRect(struct Screen *, struct Visible *);
void sav_ClearWindow(struct Window *);

// tags
BOOL sav_TagInTagList(ULONG *, struct TagItem *);
BOOL sav_TagNotInTagList(ULONG *, struct TagItem *);
struct TagItem *sav_CloneAcceptTagList(struct TagItem *, ULONG *);

// speicher
APTR sav_AllocVec(ULONG, ULONG);
void sav_FreeVec(APTR);
//void sav_DisposeMemList(void);
long sav_FreeTotalMem(void);
long sav_FreeChipMem(void);
long sav_FreeFastMem(void);
APTR sav_CloneMem(APTR, long);

// listen
void sav_InitListHeader(struct List *);
struct Node *sav_GetHead(struct List *);
struct Node *sav_GetTail(struct List *);
struct Node *sav_GetSucc(struct Node *);
struct Node *sav_GetPred(struct Node *);
long sav_ListLen(struct List *);
long sav_TailLen(struct Node *);
long sav_NodeNum(struct List *, struct Node *);
struct Node *sav_FindNodeNum(struct List *, long);
BOOL sav_IsListEmpty(struct List *);
BOOL sav_ForAllNodes(struct List *, BOOL (*)(struct Node *, APTR), APTR);
void sav_NodeToHead(struct List *, struct Node *);
void sav_NodeToTail(struct List *, struct Node *);
void sav_NodeUp(struct List *, struct Node *);
void sav_NodeDown(struct List *, struct Node *);
void sav_SwapNodes(struct Node *, struct Node *);

// system
void sav_GetSysInfo(struct SysInfo *);
long sav_TotalStackSize(void);
long sav_StackLeft(void);
BOOL sav_SetCustomStack(long);
void sav_StackSwap(struct StackSwapStruct *);
void sav_DisposeCustomStack(void);
int sav_KickVersion(void);
APTR sav_LockWindow(struct Window *);
void sav_UnlockWindow(APTR);
int sav_GetSeed(void);
struct Library *sav_FindLibrary(char *);
struct Device *sav_FindDevice(char *);
long sav_ConvertVerString(char *);
long sav_GetReg(long);
BOOL sav_ExistsDosEntry(char *, long);
void sav_PushRegs(void);
void sav_PopRegs(void);
struct Window *sav_QuietProcess(void);
void sav_KPrint(char *);
BOOL sav_OpenCloseLibrary(char *);
APTR sav_ReadSP(void);
void sav_WriteSP(APTR);
long sav_PutMsg(char *, struct Message *);

// logic
long sav_BitAND(long, long);
long sav_BitOR(long, long);
long sav_BitXOR(long, long);
long sav_BitNOT(long);
long sav_BitValue(long, long);
long sav_CheckBitPattern(long, long);
long sav_BitChange(long, long);

// sonstiges
BOOL sav_Even(long);
void sav_PushA4(void);
void sav_PopA4(void);
void sav_TempA4(void);
void sav_UntempA4(void);
void sav_PutChar(APTR, LONG);
void sav_PutInt(APTR, LONG);
void sav_PutLong(APTR, LONG);
char sav_GetChar(APTR);
int sav_GetInt(APTR);
long sav_GetLong(APTR);
void sav_Beep(void);
long sav_EasyRequest(struct Window *, char *, char *, char *, APTR);
void sav_InitTAttr(struct TextAttr *, struct TextFont *);
ULONG sav_TextLength(char *, struct TextFont *);
APTR sav_StringF(char *, APTR);
APTR sav_StringF2(char *, APTR);
long sav_RangeNumber(long, long, long);

#endif

