/*
**  ExtFrameIClass - (c) 1999 Matteo Cavalleri
**
**
**  Please read the documentation for
**  usage and license of this source.
**
**  TODO
**
**        - bordi di spessore diverso
**        - support disabled state?
**
**
**  HISTORY
**
**    0.6 - Some small enanchements.
**
**    0.5 - Now this class is a subclass of the imageclass and is
**          a complete replacement of the frameiclass (needs lot
**          of rework though...).
**        - Changed again some variables type.
**        - For compatibility reason the normal and selected frame
**          are no longer indipendent as in the previous versions,
**          moreover this doubled arrays number... :-(
**        - Again for compatibility reason all the attributes name
**          and attributes values have changed.
**        - Now when no custom pens are supplied with the appropriate
**          tags the class use drawinfo pens (like frameiclass does).
**          This added a lot of if() and a new pen array... :-(
**        - Fixed a "visual" bug. Now selected frames should be
**          identical to the standard frameiclass frames.
**        - changed "Spacing" in "Distance".
**
**    0.4 - Added EFA_InnerSpacing. This required a little rework
**          of some method but increased a bit class flexibility.
**        - Fixed a stupid bug.
**        - IA_EdgesOnly can't be read with OM_GET, so I must
**          handle it by myself. :(
**        - Changed a little OM_NEW. This should make tag reading
**          a little faster (and it made the class smaller). :-)
**          Thanks to Tim Stack and his html boopsi pages for this
**          tip. ;-)
**        - Changed a couple of variables types. Made the code
**          a little smaller, but how faster or slower?
**
**    0.3 - Added PEN_HALFSHINE and PEN_SELBACKGROUND.
**        - At the end we have a decent XEN frame. :)
**        - I misunderstood the meaning of FRAMEF_SPECIFY, now
**          it should work the right way...
**        - Added the flag FRAMEF_QUERY. See the docs for the
**          purpose of this flag (Requested by Bernardo Innocenti).
**
**    0.2 - Changed EF_#? defines to EFRAMES_#? to avoid
**          confusion with the frameiclass tags.
**        - Normal and selected frames are now indipendently
**          selectable. Added some preset.
**        - fixed a little bug in the IM_ERASE method.
**        - in the previous version was possible to set non
**          existing frames. fixed.
**        - changed a couple of variables type. I hope this
**          will avoid a lot of conversions from ULONG to
**          WORD, UWORD to WORD, UWORD to LONG, etc. (at
**          least it seems to have reduced the object size)
**
**    0.1 - First version.
*/

/* /// "Include" */

/* Avoid including this header file because of
** conflicting definitions in BoopsiStubs.h
*/
#define  CLIB_ALIB_PROTOS_H

#include <intuition/intuition.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/imageclass.h>
#include <intuition/screens.h>
#include <graphics/rastport.h>
#include <utility/tagitem.h>

#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/utility.h>

#include <BoopsiStubs.h>
#include <BoopsiLib.h>
#include <images/ExtFrameIClass.h>


/* /// */
/* /// "Global" */

/* library version we need */
#ifdef OS30_ONLY
    #define WANTEDVER      39
#else
    #define WANTEDVER      37
#endif

/* frame colors */
#define PEN_BACKGROUND      0
#define PEN_SELBACKGROUND   1
#define PEN_SHINE           2
#define PEN_SHADOW          3
#define PEN_HALFSHINE       4
#define PEN_HALFSHADOW      5

/* number of color used */
#define EXTFRAME_MAXPEN     6

/* this just makes the source a little bit more readable */
#define PEN(p)      (data->pen[p] == -1 ? data->defpen[p] : data->pen[p])
/* this is a little variant needed only for EF_EraseFrame() */
#define BGPEN(p)    (data->pen[p] == -1 ? msg->imp_RPort->BgPen : data->pen[p])

/* the following arrays should avoid a lot of if() and
** switch() and easily let us add new frames types
** without modifying too much the code of the methods
*/

/* frame thickness, usually 1, 2 or 3.
** afaik other values doesn't make sense
*/
static const WORD thickness[] =
{
    1,                  /* FRAME_DEFAULT     */
    2,                  /* FRAME_BUTTON      */
    2,                  /* FRAME_RIDGE       */
    3,                  /* FRAME_ICONDROPBOX */
    1,                  /* FRAME_FLAT        */
    2,                  /* FRAME_XEN         */
    2,                  /* FRAME_BUTTON2     */
    2                   /* FRAME_RIDGE2      */
};

/* which frames should be rendered with 'rounded'
** corners. this is currently needed for FRAME_FLAT
** and FRAME_XEN, since they would look really ugly
*/
static const WORD skip_pixel[] =
{
    1,                  /* FRAME_DEFAULT     */
    1,                  /* FRAME_BUTTON      */
    1,                  /* FRAME_RIDGE       */
    1,                  /* FRAME_ICONDROPBOX */
    0,                  /* FRAME_FLAT        */
    0,                  /* FRAME_XEN         */
    1,                  /* FRAME_BUTTON2     */
    1                   /* FRAME_RIDGE2      */
};

/* pens to be used for the various frames */

/* upper and left borders, normal frame */
static const UWORD pen_topleft1_n[] =
{
    PEN_SHINE,          /* FRAME_DEFAULT     */
    PEN_SHINE,          /* FRAME_BUTTON      */
    PEN_SHINE,          /* FRAME_RIDGE       */
    PEN_SHINE,          /* FRAME_ICONDROPBOX */
    PEN_SHADOW,         /* FRAME_FLAT        */
    PEN_SHADOW,         /* FRAME_XEN         */
    PEN_SHINE,          /* FRAME_BUTTON2     */
    PEN_SHINE           /* FRAME_RIDGE2      */
};

/* right and bottom borders, normal frame */
static const UWORD pen_bottomright1_n[] =
{
    PEN_SHADOW,         /* FRAME_DEFAULT     */
    PEN_SHADOW,         /* FRAME_BUTTON      */
    PEN_SHADOW,         /* FRAME_RIDGE       */
    PEN_SHADOW,         /* FRAME_ICONDROPBOX */
    PEN_SHADOW,         /* FRAME_FLAT        */
    PEN_SHADOW,         /* FRAME_XEN         */
    PEN_SHADOW,         /* FRAME_BUTTON2     */
    PEN_SHADOW          /* FRAME_RIDGE2      */
};

/* upper and left borders, normal frame */
static const UWORD pen_topleft2_n[] =
{
    NULL,               /* FRAME_DEFAULT     */
    PEN_SHINE,          /* FRAME_BUTTON      */
    PEN_SHADOW,         /* FRAME_RIDGE       */
    PEN_SHADOW,         /* FRAME_ICONDROPBOX */
    NULL,               /* FRAME_FLAT        */
    PEN_SHINE,          /* FRAME_XEN         */
    PEN_SHINE,          /* FRAME_BUTTON2     */
    PEN_SHADOW          /* FRAME_RIDGE2      */
};

/* right and bottom borders, normal frame */
static const UWORD pen_bottomright2_n[] =
{
    NULL,               /* FRAME_DEFAULT     */
    PEN_SHADOW,         /* FRAME_BUTTON      */
    PEN_SHINE,          /* FRAME_RIDGE       */
    PEN_SHINE,          /* FRAME_ICONDROPBOX */
    NULL,               /* FRAME_FLAT        */
    PEN_HALFSHADOW,     /* FRAME_XEN         */
    PEN_SHADOW,         /* FRAME_BUTTON2     */
    PEN_SHINE           /* FRAME_RIDGE2      */
};

/* upper and left borders, selected frame */
static const UWORD pen_topleft1_s[] =
{
    PEN_SHADOW,         /* FRAME_DEFAULT     */
    PEN_SHADOW,         /* FRAME_BUTTON      */
    PEN_SHADOW,         /* FRAME_RIDGE       */
    PEN_SHADOW,         /* FRAME_ICONDROPBOX */
    PEN_SHINE,          /* FRAME_FLAT        */
    PEN_SHADOW,         /* FRAME_XEN         */
    PEN_SHADOW,         /* FRAME_BUTTON2     */
    PEN_SHINE           /* FRAME_RIDGE2      */
};

/* right and bottom borders, selected frame */
static const UWORD pen_bottomright1_s[] =
{
    PEN_SHINE,          /* FRAME_DEFAULT     */
    PEN_SHINE,          /* FRAME_BUTTON      */
    PEN_SHINE,          /* FRAME_RIDGE       */
    PEN_SHINE,          /* FRAME_ICONDROPBOX */
    PEN_SHINE,          /* FRAME_FLAT        */
    PEN_SHADOW,         /* FRAME_XEN         */
    PEN_SHINE,          /* FRAME_BUTTON2     */
    PEN_SHADOW          /* FRAME_RIDGE2      */
};

/* upper and left borders, selected frame */
static const UWORD pen_topleft2_s[] =
{
    NULL,               /* FRAME_DEFAULT     */
    PEN_SHADOW,         /* FRAME_BUTTON      */
    PEN_SHINE,          /* FRAME_RIDGE       */
    PEN_SHINE,          /* FRAME_ICONDROPBOX */
    NULL,               /* FRAME_FLAT        */
    PEN_BACKGROUND,     /* FRAME_XEN         */
    PEN_SHINE,          /* FRAME_BUTTON2     */
    PEN_SHINE           /* FRAME_RIDGE2      */
};

/* right and bottom borders, selected frame */
static const UWORD pen_bottomright2_s[] =
{
    NULL,               /* FRAME_DEFAULT     */
    PEN_SHINE,          /* FRAME_BUTTON      */
    PEN_SHADOW,         /* FRAME_RIDGE       */
    PEN_SHADOW,         /* FRAME_ICONDROPBOX */
    NULL,               /* FRAME_FLAT        */
    PEN_SELBACKGROUND,  /* FRAME_XEN         */
    PEN_SHADOW,         /* FRAME_BUTTON2     */
    PEN_SHADOW          /* FRAME_RIDGE2      */
};

/* these are our flags */
#define EDGESONLY       (1 << 0)
#define RECESSED        (1 << 1)
//#define SUPPORTDISABLE  (1 << 2)

/* our private data */
struct EF_Data
{
    ULONG  frame_type;
    WORD   frame_distance;           /* space between the frame and the gadget content            */
    WORD   inner_distance;           /* space between the two borders of double frames            */

    WORD   pen[EXTFRAME_MAXPEN];    /* our custom pens                                           */
    WORD   defpen[EXTFRAME_MAXPEN]; /* remapped drawinfo pens (when no custom pens are supplied) */

    UWORD  flags;
};

#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)
    const UBYTE LibName[] = NAME;
    const UBYTE LibVer[] = { '$', 'V', 'E', 'R', ':', ' ' };
    const UBYTE LibId[] = NAME " " VERSION " " DATE " 1999 Matteo Cavalleri\n";

    static struct ExecBase      *SysBase = NULL;
    static struct IntuitionBase *IntuitionBase = NULL;
    static struct GfxBase       *GfxBase = NULL;
    static struct UtilityBase   *UtilityBase = NULL;
#endif

/* /// */
/* /// "Prototyping" */

static ULONG HOOKCALL EF_Dispatcher(REG(a0, Class *cl), REG(a2, struct Image *img), REG(a1, Msg msg));

static ULONG EF_DrawFrame(Class *cl, struct Image *img, struct impDraw *msg);
static ULONG EF_EraseFrame(Class *cl, struct Image *img, struct impErase *msg);
static ULONG EF_FrameBox(Class *cl, struct Image *img, struct impFrameBox *msg);
static ULONG EF_Set(Class *cl, struct Image *img, struct opSet *msg);
static ULONG EF_Get(Class *cl, struct Image *img, struct opGet *msg);
static struct Image *EF_New(Class *cl, struct Image *img, struct opSet *msg);

#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)
    /* class library support functions      */
    struct ClassLibrary *HOOKCALL _UserLibInit(REG(a6, struct ClassLibrary *mybase));
    struct ClassLibrary *HOOKCALL _UserLibCleanup(REG(a6, struct ClassLibrary *mybase));
    Class *HOOKCALL _GetEngine(REG(a6, struct ClassLibrary *mybase));
#endif

/* /// */

/* /// "EF_Dispatcher" */

static ULONG HOOKCALL EF_Dispatcher(REG(a0, Class *cl), REG(a2, struct Image *img), REG(a1, Msg msg))
{
    switch(msg->MethodID)
    {
        case IM_DRAW:
        case IM_DRAWFRAME:
            return EF_DrawFrame(cl, img, (struct impDraw *)msg);

        case IM_ERASE:
        case IM_ERASEFRAME:
            return EF_EraseFrame(cl, img, (struct impErase *)msg);

        /* This is probably wrong. I still
        ** need to remember to have a look
        ** at the RKM :-)
        */
        case IM_HITTEST:
        case IM_HITFRAME:
        {
            WORD right, bottom;

            /* calculate our size */
            right  = img->LeftEdge + ((msg->MethodID == IM_HITTEST) ? img->Width :
                            ((struct impHitTest *)msg)->imp_Dimensions.Width) - 1;

            bottom = img->TopEdge + ((msg->MethodID == IM_HITTEST) ? img->Height :
                           ((struct impHitTest *)msg)->imp_Dimensions.Height) - 1;

            /* is the point within us? */
            if((((struct impHitTest *)msg)->imp_Point.X >= img->LeftEdge) &&
               (((struct impHitTest *)msg)->imp_Point.X <= right) &&
               (((struct impHitTest *)msg)->imp_Point.Y >= img->TopEdge) &&
               (((struct impHitTest *)msg)->imp_Point.Y <= bottom))
               return 1;

            return 0;
        }

        case IM_FRAMEBOX:
            return EF_FrameBox(cl, img, (struct impFrameBox *)msg);

        case OM_SET:
//        {
//            ULONG result;

            EF_Set(cl, img, (struct opSet *)msg);
            /* we put DoSuperMethodA() here because
            ** we use EF_Set for our purposes too
            */
            return DoSuperMethodA(cl, (Object *)img, msg);

//            result  = EF_Set(cl, img, (struct opSet *)msg);
//            result += DoSuperMethodA(cl, (Object *)img, msg);
//            return result;
//        }

        case OM_GET:
            return EF_Get(cl, img, (struct opGet *)msg);

        case OM_NEW:
            return (ULONG)EF_New(cl, img, (struct opSet *)msg);

        default:
            return DoSuperMethodA(cl, (Object *)img, msg);
    }
}

/* /// */

/* /// "EF_DrawFrame" */

static ULONG EF_DrawFrame(Class *cl, struct Image *img, struct impDraw *msg)
{
    struct EF_Data *data;
    WORD   left, top, right, bottom;
    ULONG  pen1, pen2;
    WORD   temp;

    /* get the address of our data */
    data = (struct EF_Data *)INST_DATA(cl, img);

    /* fill the defpen array (i think it's better to
    ** do this always since a single frame may be
    ** used to draw itself not only around different
    ** gadgets but maybe even on different screens)
    */
    if(msg->imp_DrInfo == NULL)
    {
        /* we may have a NULL drawinfo! */
        data->defpen[PEN_BACKGROUND] = 0;
        data->defpen[PEN_SELBACKGROUND] = 3;
        data->defpen[PEN_SHINE] = 2;
        data->defpen[PEN_SHADOW] = 1;
        data->defpen[PEN_HALFSHINE] = 2;
        data->defpen[PEN_HALFSHADOW] = 1;
    }
    else
    {
        /* screens.h says that BACKGROUNDPEN is always 0,
        ** but VisualPrefs can set another pen for it, so
        ** i think it's safer to read it from DrInfo anyway.
        */
        data->defpen[PEN_BACKGROUND] = msg->imp_DrInfo->dri_Pens[BACKGROUNDPEN];
        data->defpen[PEN_SELBACKGROUND] = msg->imp_DrInfo->dri_Pens[FILLPEN];
        data->defpen[PEN_SHINE] = msg->imp_DrInfo->dri_Pens[SHINEPEN];
        data->defpen[PEN_SHADOW] = msg->imp_DrInfo->dri_Pens[SHADOWPEN];
        data->defpen[PEN_HALFSHINE] = msg->imp_DrInfo->dri_Pens[SHINEPEN];
        data->defpen[PEN_HALFSHADOW] = msg->imp_DrInfo->dri_Pens[SHADOWPEN];
    }

    /* just some thoughts on the defpen array:
    **
    ** since most of the class' pens are just equivalent of
    ** intuition pens we may create a pen array with the same
    ** pen indexes of the drawinfo pen array. this would avoid
    ** the defpen array and let use the drawinfo array directly,
    ** but since it doesn't have PEN_HALFSHINE and PEN_HALFSHADOW
    ** equivalents (which are remapped to SHINEPEN and SHADOWPEN)
    ** this is not possible unfortunately. we may however give
    ** PEN_HALFSHINE and PEN_HALFSHADOW the same index of other
    ** pen like e.g. DETAILPEN but i think this would lead to
    ** unpredictable result when the user use VisualPrefs.
    */

    /* calculate our size */
    left   = msg->imp_Offset.X;
    top    = msg->imp_Offset.Y;
    right  = msg->imp_Offset.X + ((msg->MethodID == IM_DRAW) ? img->Width : msg->imp_Dimensions.Width) - 1;
    bottom = msg->imp_Offset.Y + ((msg->MethodID == IM_DRAW) ? img->Height : msg->imp_Dimensions.Height) - 1;

/* this little optimization can be done only
** with version 3.0 of the operating system
*/
#ifdef OS30_ONLY
    /* should we fill ourselves? */
    if((data->flags & EDGESONLY) == FALSE)
    {
        switch(msg->imp_State)
        {
            case IDS_SELECTED:
            case IDS_INACTIVESELECTED:
            case IDS_SELECTEDDISABLED:
                pen1 = PEN(PEN_SELBACKGROUND);
                break;
            default:
                pen1 = PEN(PEN_BACKGROUND);
                break;
        }
        SetABPenDrMd(msg->imp_RPort, pen1, 0, JAM1);
        RectFill(msg->imp_RPort, left + 1, top + 1, right - 1, bottom - 1);
    }
    else
#endif

    /* we must set the drawing mode since the rastport we
    ** are going to use is shared with other tasks.
    ** since SetDrMd() is slow (it has to calculate blitter
    ** minterm) we call it only when it is necessary.
    */
    if(GetDrMd(msg->imp_RPort) != JAM1) SetDrMd(msg->imp_RPort, JAM1);

#ifndef OS30_ONLY
    /* should we fill ourselves? */
    if((data->flags & EDGESONLY) == FALSE)
    {
        switch(msg->imp_State)
        {
            case IDS_SELECTED:
            case IDS_INACTIVESELECTED:
            case IDS_SELECTEDDISABLED:
                pen1 = PEN(PEN_SELBACKGROUND);
                break;
            default:
                pen1 = PEN(PEN_BACKGROUND);
                break;
        }
        SetAPen(msg->imp_RPort, pen1);
        RectFill(msg->imp_RPort, left + 1, top + 1, right - 1, bottom - 1);
    }
#endif

    /* start drawing the frame */

    /* NB: for thick frames now we are going
    ** to draw the outer part of the frame.
    */
    switch(msg->imp_State)
    {
        case IDS_SELECTED:
        case IDS_INACTIVESELECTED:
            if((data->flags & RECESSED) && (data->frame_type != FRAME_XEN))
            {
                pen1 = PEN(pen_topleft1_n[data->frame_type]);
                pen2 = PEN(pen_bottomright1_n[data->frame_type]);
                break;
            }
            pen1 = PEN(pen_topleft1_s[data->frame_type]);
            pen2 = PEN(pen_bottomright1_s[data->frame_type]);
            break;
//        case IDS_DISABLED:
//        case IDS_INACTIVEDISABLED:
//        case IDS_SELECTEDDISABLED:
//            break;
        default:
            if((data->flags & RECESSED) && (data->frame_type != FRAME_XEN))
            {
                pen1 = PEN(pen_topleft1_s[data->frame_type]);
                pen2 = PEN(pen_bottomright1_s[data->frame_type]);
                break;
            }
            pen1 = PEN(pen_topleft1_n[data->frame_type]);
            pen2 = PEN(pen_bottomright1_n[data->frame_type]);
            break;
    }

    /* check if we must draw top left and right bottom pixels */
    temp = skip_pixel[data->frame_type];

    /* first we draw the right and bottom border */
    SetAPen(msg->imp_RPort, pen2);
    RectFill(msg->imp_RPort, right, top + temp, right, bottom);
    RectFill(msg->imp_RPort, left + temp, bottom, right, bottom);

    /* SetAPen() is a slow function, since it needs to calculate
    ** minterms everytime, so i think it's better to add this little
    ** if() statement than calling SetAPen() even when not needed
    */
    if(pen1 != pen2) SetAPen(msg->imp_RPort, pen1);
    RectFill(msg->imp_RPort, left, top, right - temp, top);
    RectFill(msg->imp_RPort, left, top, left, bottom - temp);

    /* check if we have finished */
    if(thickness[data->frame_type] == 1) return 1;

    /* calculate frame thickness */
    switch(thickness[data->frame_type])
    {
        case 3:
            left   += data->inner_distance;
            top    += data->inner_distance;
            right  -= data->inner_distance;
            bottom -= data->inner_distance;
            /* we fall down here */
        case 2:
            left++;
            top++;
            right--;
            bottom--;
            break;
    }

    /* draw the inner part for double frames. */
    switch(msg->imp_State)
    {
        case IDS_SELECTED:
        case IDS_INACTIVESELECTED:
            if((data->flags & RECESSED) && (data->frame_type != FRAME_XEN))
            {
                pen1 = PEN(pen_topleft2_n[data->frame_type]);
                pen2 = PEN(pen_bottomright2_n[data->frame_type]);
                break;
            }
            pen1 = PEN(pen_topleft2_s[data->frame_type]);
            pen2 = PEN(pen_bottomright2_s[data->frame_type]);
            break;
//        case IDS_DISABLED:
//        case IDS_INACTIVEDISABLED:
//        case IDS_SELECTEDDISABLED:
//            break;
        default:
            if((data->flags & RECESSED) && (data->frame_type != FRAME_XEN))
            {
                pen1 = PEN(pen_topleft2_s[data->frame_type]);
                pen2 = PEN(pen_bottomright2_s[data->frame_type]);
                break;
            }
            pen1 = PEN(pen_topleft2_n[data->frame_type]);
            pen2 = PEN(pen_bottomright2_n[data->frame_type]);
            break;
    }

    SetAPen(msg->imp_RPort, pen2);
    RectFill(msg->imp_RPort, right, top, right, bottom);
    RectFill(msg->imp_RPort, left, bottom, right, bottom);

    if(pen1 != pen2) SetAPen(msg->imp_RPort, pen1);
    RectFill(msg->imp_RPort, left, top, right, top);
    RectFill(msg->imp_RPort, left, top, left, bottom);

    return 1;
}

/* /// */
/* /// "EF_EraseFrame" */

static ULONG EF_EraseFrame(Class *cl, struct Image *img, struct impErase *msg)
{
    struct EF_Data *data;
    WORD   left, top, right, bottom;
    WORD   tn = 0;

    /* get the address of our data */
    data = (struct EF_Data *)INST_DATA(cl, img);

    /* calculate our size */
    left   = msg->imp_Offset.X;
    top    = msg->imp_Offset.Y;
    right  = msg->imp_Offset.X + ((msg->MethodID == IM_ERASE) ? img->Width : msg->imp_Dimensions.Width) - 1;
    bottom = msg->imp_Offset.Y + ((msg->MethodID == IM_ERASE) ? img->Height : msg->imp_Dimensions.Height) - 1;

#ifdef OS30_ONLY
    /* this little optimization can be done only
    ** with version 3.0 of the operating system
    ** NB we can avoid to use GetDrMd() since we
    ** need to call SetABPenDrMd() in any case
    */
    SetABPenDrMd(msg->imp_RPort, BGPEN(PEN_BACKGROUND), 0, JAM1);
#else
    /* we must set the drawing mode since the rastport we
    ** are going to use is shared with other tasks.
    ** since SetDrMd() is slow (it has to calculate blitter
    ** minterm) we call it only when it is necessary
    */
    if(GetDrMd(msg->imp_RPort) != JAM1) SetDrMd(msg->imp_RPort, JAM1);

    /* select the bg color (NB: we don't have any drawinfo here, so
    ** i think it's better to use msg->imp_RPort->BgPen than a
    ** simple zero since this may give us VisualPrefs compatibility)
    */
    SetAPen(msg->imp_RPort, BGPEN(PEN_BACKGROUND));
#endif

    /* should we fill ourseleves? */
    if((data->flags & EDGESONLY) == FALSE)
    {
        RectFill(msg->imp_RPort, left, top, right, bottom);
        return 1;
    }

    /* calculate correct frame size */
    switch(thickness[data->frame_type])
    {
        case 3:
            tn = data->inner_distance;
            /* we fall down here */
        case 2:
            tn++;
            /* we fall down here */
        default:
            break;
    }

    /* erase the border */
    RectFill(msg->imp_RPort, left, top, right, top + tn);
    RectFill(msg->imp_RPort, left, top, left + tn, bottom);
    RectFill(msg->imp_RPort, right - tn, top, right, bottom);
    RectFill(msg->imp_RPort, left, bottom - tn, right, bottom);

    return 1;
}

/* /// */
/* /// "EF_FrameBox" */

static ULONG EF_FrameBox(Class *cl, struct Image *img, struct impFrameBox *msg)
{
    struct EF_Data *data;
    WORD   tn;

    /* our with and height are supplied in imp_Framebox,
    ** we need just to centre ourselves over the gadget
    */
    if(msg->imp_FrameFlags & FRAMEF_SPECIFY)
    {
        msg->imp_FrameBox->Left = msg->imp_ContentsBox->Left +
            ((msg->imp_ContentsBox->Width - msg->imp_FrameBox->Width) >> 1);

        msg->imp_FrameBox->Top  = msg->imp_ContentsBox->Top +
            ((msg->imp_ContentsBox->Width - msg->imp_FrameBox->Height) >> 1);

        return 1;
    }

    /* get the address of our data */
    data = (struct EF_Data *)INST_DATA(cl, img);

    /* calculate needed space */
    tn = data->frame_distance + thickness[data->frame_type];
    /* for these frames we must consider also the inner space */
    if(thickness[data->frame_type] == 3) tn += data->inner_distance - 1;

    /* calculate the size of our contents when
    ** we have imp_FrameBox dimensions...
    */
    if(msg->imp_FrameFlags & FRAMEF_QUERY)
    {
        msg->imp_ContentsBox->Left   = msg->imp_FrameBox->Left + tn;
        msg->imp_ContentsBox->Top    = msg->imp_FrameBox->Top + tn;
        msg->imp_ContentsBox->Width  = msg->imp_FrameBox->Width - tn - tn;
        msg->imp_ContentsBox->Height = msg->imp_FrameBox->Height - tn - tn;

        return 1;
    }

    /* set our size */
    msg->imp_FrameBox->Left   = msg->imp_ContentsBox->Left - tn;
    msg->imp_FrameBox->Top    = msg->imp_ContentsBox->Top - tn;
    msg->imp_FrameBox->Width  = msg->imp_ContentsBox->Width + tn + tn;
    msg->imp_FrameBox->Height = msg->imp_ContentsBox->Height + tn + tn;

    return 1;
}

/* /// */
/* /// "EF_Set" */

static ULONG EF_Set(Class *cl, struct Image *img, struct opSet *msg)
{
    struct EF_Data *data;
    struct TagItem *tag;

    /* get the address of our data */
    data = (struct EF_Data *)INST_DATA(cl, img);

    /* scan the tag list */
    while((tag = NextTagItem(&(msg->ops_AttrList))))
    {
        switch(tag->ti_Tag)
        {
            case IA_FrameType:
                /* avoid setting non existing frames */
                data->frame_type = (tag->ti_Data > EXTFRAME_MAXFRAME ? FRAME_DEFAULT : tag->ti_Data);
                break;

            case IA_FrameDistance:
                data->frame_distance = tag->ti_Data;
                break;
            case IA_InnerDistance:
                data->inner_distance = tag->ti_Data;
                break;

            case IA_PenBackground:
                data->pen[PEN_BACKGROUND] = tag->ti_Data;
                break;
            case IA_PenSelBackground:
                data->pen[PEN_SELBACKGROUND] = tag->ti_Data;
                break;
            case IA_PenShine:
                data->pen[PEN_SHINE] = tag->ti_Data;
                break;
            case IA_PenShadow:
                data->pen[PEN_SHADOW] = tag->ti_Data;
                break;
            case IA_PenHalfShine:
                data->pen[PEN_HALFSHINE] = tag->ti_Data;
                break;
            case IA_PenHalfShadow:
                data->pen[PEN_HALFSHADOW] = tag->ti_Data;
                break;

            case IA_EdgesOnly:
                if(tag->ti_Data == FALSE) data->flags &= ~EDGESONLY; else data->flags |= EDGESONLY;
                break;
            case IA_Recessed:
                if(tag->ti_Data == FALSE) data->flags &= ~RECESSED; else data->flags |= RECESSED;
                break;
//            case IA_SupportsDisable:
//                if(tag->ti_Data == FALSE) data->flags &= ~SUPPORTDISABLE; else data->flags |= SUPPORTDISABLE;
//                break;

            default:
                break;
        }
    }

    /* we call our superclass method in the dispatcher function */
    return 1;
}

/* /// */
/* /// "EF_Get" */

static ULONG EF_Get(Class *cl, struct Image *img, struct opGet *msg)
{
    struct EF_Data *data;

    /* get the address of our data */
    data = (struct EF_Data *)INST_DATA(cl, img);

    switch(msg->opg_AttrID)
    {
        case IA_FrameType:
            *(msg->opg_Storage) = (ULONG)data->frame_type;
            break;

        case IA_FrameDistance:
            *(msg->opg_Storage) = (ULONG)data->frame_distance;
            break;
        case IA_InnerDistance:
            *(msg->opg_Storage) = (ULONG)data->inner_distance;
            break;

        case IA_PenBackground:
            *(msg->opg_Storage) = (ULONG)data->pen[PEN_BACKGROUND];
            break;
        case IA_PenSelBackground:
            *(msg->opg_Storage) = (ULONG)data->pen[PEN_SELBACKGROUND];
            break;
        case IA_PenShine:
            *(msg->opg_Storage) = (ULONG)data->pen[PEN_SHINE];
            break;
        case IA_PenShadow:
            *(msg->opg_Storage) = (ULONG)data->pen[PEN_SHADOW];
            break;
        case IA_PenHalfShine:
            *(msg->opg_Storage) = (ULONG)data->pen[PEN_HALFSHINE];
            break;
        case IA_PenHalfShadow:
            *(msg->opg_Storage) = (ULONG)data->pen[PEN_HALFSHADOW];
            break;

        case IA_EdgesOnly:
            *(msg->opg_Storage) = (data->flags & EDGESONLY ? TRUE : FALSE);
            break;
        case IA_Recessed:
            *(msg->opg_Storage) = (data->flags & RECESSED ? TRUE : FALSE);
            break;
//        case IA_SupportsDisable:
//            *(msg->opg_Storage) = (data->flags & SUPPORTDISABLE ? TRUE : FALSE);
//            break;

        default:
            return DoSuperMethodA(cl, (Object *)img, (Msg)msg);
    }
    return 1;
}

/* /// */
/* /// "EF_New" */

static struct Image *EF_New(Class *cl, struct Image *img, struct opSet *msg)
{
    struct Image *newimg;
    struct EF_Data *data;

    newimg = (struct Image *)DoSuperMethodA(cl, (Object *)img, (Msg)msg);
    if(newimg != NULL)
    {
        /* get the address of our data */
        data = (struct EF_Data *)INST_DATA(cl, newimg);

        /* set up default attribute values */
        data->frame_type = FRAME_DEFAULT;
        data->frame_distance = 2;
        data->inner_distance = 1;
        data->pen[PEN_BACKGROUND] = -1;
        data->pen[PEN_SELBACKGROUND] = -1;
        data->pen[PEN_SHINE] = -1;
        data->pen[PEN_SHADOW] = -1;
        data->pen[PEN_HALFSHINE] = -1;
        data->pen[PEN_HALFSHADOW] = -1;
        data->flags = NULL;

        /* read the various attributes */
        EF_Set(cl, newimg, (struct opSet *)msg);
    }
    return newimg;
}

/* /// */

/* /// "makeExtFrameIClass" */

#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)
    /* use this function to create a private class instead of a public one */
    Class *makeExtFrameIClass(void)
    {
        Class *cl;

        cl = MakeClass(NULL, IMAGECLASS, NULL, sizeof(struct EF_Data), 0);
        if(cl != NULL) cl->cl_Dispatcher.h_Entry = (ULONG (*)())EF_Dispatcher;

        return cl;
    }
#endif

/* /// */

/* /// "_UserLibInit" */

#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)

/* use this function to create a public class instead of a private one */
struct ClassLibrary *HOOKCALL _UserLibInit(REG(a6, struct ClassLibrary *mybase))
{
    /* initialize sysbase */
    SysBase = *((struct ExecBase **)4);

    /* open needed libraries */
    IntuitionBase = (struct IntuitionBase *)OpenLibrary ("intuition.library", WANTEDVER);
    GfxBase = (struct GfxBase *)OpenLibrary ("graphics.library", WANTEDVER);
    UtilityBase = (struct UtilityBase *)OpenLibrary ("utility.library", WANTEDVER);

    if(IntuitionBase && GfxBase && UtilityBase)
    {
        /* create our public classes */
        mybase->cl_Class = MakeClass(EXTFRAMEICLASS, IMAGECLASS, NULL, sizeof(struct EF_Data), 0);
        if(mybase->cl_Class != NULL)
        {
            mybase->cl_Class->cl_Dispatcher.h_Entry = (ULONG (*)())EF_Dispatcher;
            /* this is for compatibility reasons, just to be sure... */
            /* mybase->cl_Class->cl_UserData = (ULONG)mybase;        */

            /* make the class public */
            AddClass(mybase->cl_Class);

            return mybase;
        }
    }

    /* something failed */
    _UserLibCleanup(mybase);
    return NULL;
}
/* /// */
/* /// "_UserLibCleanup" */

struct ClassLibrary *HOOKCALL _UserLibCleanup(REG(a6, struct ClassLibrary *mybase))
{
    if(mybase->cl_Class)
        if(!FreeClass(mybase->cl_Class)) return NULL;

    CloseLibrary((struct Library *)UtilityBase);
    CloseLibrary((struct Library *)GfxBase);
    CloseLibrary((struct Library *)IntuitionBase);

    return mybase;
}

/* /// */
/* /// "_GetEngine" */

Class *HOOKCALL _GetEngine(REG(a6, struct ClassLibrary *mybase))
{
    return mybase->cl_Class;
}

#endif

/* /// */

