/*
**	$Id: VectorGlyphIClass.c,v 1.2 1999/02/07 14:32:00 bernie Exp $
**
**	Copyright (C) 1995,96,97,98 Bernardo Innocenti
**	All rights reserved.
**
**	Use 4 chars wide TABs to read this file
**
**	"vectorglyphiclass", a vector image class built
**	on top of the "imageclass", providing some useful
**	glyphs for buttons.
*/

#define USE_BUILTIN_MATH
#define INTUI_V36_NAMES_ONLY
#define INTUITION_IOBSOLETE_H
#define __USE_SYSBASE
#define  CLIB_ALIB_PROTOS_H		/* Avoid dupe defines of boopsi funcs */

#include <exec/types.h>
#include <exec/memory.h>
#include <utility/tagitem.h>
#include <graphics/gfxbase.h>
#include <intuition/intuition.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/imageclass.h>

#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <proto/utility.h>

#include <CompilerSpecific.h>
#include <DebugMacros.h>
#include <BoopsiStubs.h>
#include <BoopsiLib.h>

#include "VectorGlyphIClass.h"

#ifdef __STORM__
	#pragma header
#endif


/* Local function prototypes */

static ULONG HOOKCALL VectorGlyphDispatcher (
	REG(a0, Class *cl),
	REG(a2, struct Image *im),
	REG(a1, Msg msg));

static ULONG	VG_IMDraw			(Class *cl, struct Image *im, struct impDraw *msg);
static ULONG	VG_OMSet			(Class *cl, struct Image *im, struct opSet *msg);
static ULONG	VG_OMNew			(Class *cl, struct Image *im, struct opSet *msg);
static void		VG_OMDispose		(Class *cl, struct Image *im, Msg msg);

static struct BitMap *CreateGlyph	(ULONG which, ULONG width, ULONG height, struct BitMap *friend);

static void	DrawPlayImage			(struct RastPort *rp, UWORD width, UWORD heigth);
static void	DrawStopImage			(struct RastPort *rp, UWORD width, UWORD heigth);
static void	DrawForwardImage		(struct RastPort *rp, UWORD width, UWORD heigth);
static void	DrawRewindImage			(struct RastPort *rp, UWORD width, UWORD heigth);
static void	DrawPickImage			(struct RastPort *rp, UWORD width, UWORD height);
static void	DrawUpArrowImage		(struct RastPort *rp, UWORD width, UWORD height);
static void	DrawDownArrowImage		(struct RastPort *rp, UWORD width, UWORD height);
static void	DrawLeftArrowImage		(struct RastPort *rp, UWORD width, UWORD height);
static void	DrawRightArrowImage		(struct RastPort *rp, UWORD width, UWORD height);



#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)

	/* Class library support functions */

	struct ClassLibrary	*	HOOKCALL _UserLibInit		(REG(a6, struct ClassLibrary *mybase));
	struct ClassLibrary	*	HOOKCALL _UserLibCleanup	(REG(a6, struct ClassLibrary *mybase));
	Class *					HOOKCALL _GetEngine			(REG(a6, struct ClassLibrary *mybase));

	/* Library data */

	const UBYTE LibName[] = "vectorglyph.image";
	const UBYTE LibVer[] = { '$', 'V', 'E', 'R', ':', ' ' };
	const UBYTE LibId[] = "vectorglyph.image 2.0 (12.12.98)  1997,98 Bernardo Innocenti\n";

	/* Workaround a bug in StormC header file <proto/utility.h> */
	#ifdef __STORM__
		#define UTILITYBASETYPE struct Library
	#else
		#define UTILITYBASETYPE struct UtilityBase
	#endif

	/* Library bases */
	static struct ExecBase		*SysBase		= NULL;
	static struct IntuitionBase	*IntuitionBase	= NULL;
	static struct GfxBase		*GfxBase		= NULL;
	static UTILITYBASETYPE		*UtilityBase	= NULL;
#endif



/* Array of function pointers to draw the various glyphs */

static void (*DrawFuncs[VG_IMGCOUNT])(struct RastPort *, UWORD, UWORD) =
{
	DrawPlayImage,		/* VG_PLAY			*/
	DrawStopImage,		/* VG_STOP			*/
	DrawForwardImage,	/* VG_REW			*/
	DrawRewindImage,    /* VG_FWD			*/
	DrawPickImage,      /* VG_PICK			*/
	DrawUpArrowImage,	/* VG_UPARROW		*/
	DrawDownArrowImage,	/* VG_DOWNARROW		*/
	DrawLeftArrowImage,	/* VG_LEFTARROW		*/
	DrawRightArrowImage	/* VG_RIGHTARROW	*/
};



static ULONG HOOKCALL VectorGlyphDispatcher(
	REG(a0, Class *cl),
	REG(a2, struct Image *im),
	REG(a1, Msg msg))

/* VectorGlyph class dispatcher - Handles all supported methods */
{
	switch (msg->MethodID)
	{
		case IM_DRAW:
		case IM_DRAWFRAME:
			return VG_IMDraw(cl, im, (struct impDraw *)msg);

		case OM_SET:
			return VG_OMSet(cl, im, (struct opSet *)msg);

		case OM_NEW:
			return VG_OMNew(cl, im, (struct opSet *)msg);

		case OM_DISPOSE:
			VG_OMDispose(cl, im, msg);
			return 0;

		default:
			DB2(DBPRINTF("VectorGlyph: passing unknown method 0x%lx to superclass\n",
				msg->MethodID);)

			/* Unsupported method: let our superclass's
			 * dispatcher take a look at it.
			 */
			return DoSuperMethodA(cl, (Object *)im, msg);
	}
}

static ULONG VG_IMDraw(UNUSED(Class *cl), struct Image *im, struct impDraw *msg)
{
	DB2 ((msg->MethodID == IM_DRAW) ?
		DBPRINTF ("VectorGlyph: IM_DRAW\n") :
		DBPRINTF ("VectorGlyph: IM_DRAWFRAME\n");)

	if ((msg->MethodID == IM_DRAWFRAME) &&
		((msg->imp_Dimensions.Width != im->Width) ||
			(msg->imp_Dimensions.Height != im->Height)))
	{
		/* Destroy current image */
		if (im->ImageData)
		{
			FreeBitMap((struct BitMap *)im->ImageData);
			im->ImageData = NULL;
		}
		im->Width = msg->imp_Dimensions.Width;
		im->Height = msg->imp_Dimensions.Height;
	}

	if (!im->ImageData)
	{
		/* Store the BitMap pointer in Image->ImageData for later use */
		im->ImageData =
			(UWORD *)CreateGlyph(im->PlanePick, im->Width, im->Height,
				msg->imp_RPort->BitMap);
	}

	if (im->ImageData)
		BltBitMapRastPort((struct BitMap *)im->ImageData,
			0, 0, msg->imp_RPort,
			msg->imp_Offset.X, msg->imp_Offset.Y, im->Width, im->Height,
			(msg->imp_State == IDS_SELECTED) ? 0x030 : 0x0C0);

	return TRUE;
}



static ULONG VG_OMSet(Class *cl, struct Image *im, struct opSet *msg)
{
	struct TagItem	*ti,
					*tstate	= msg->ops_AttrList;
	BOOL bitmap_invalid = FALSE;

	DB2(DBPRINTF("VectorGlyph: OM_SET\n");)
	ASSERT_VALID_PTR(tstate)

	while ((ti = NextTagItem(&tstate)))
		switch (ti->ti_Tag)
		{
			case IA_Width:
				if (im->Width != (WORD)ti->ti_Data)
					bitmap_invalid = TRUE;
				break;

			case IA_Height:
				if (im->Height != (WORD)ti->ti_Data)
					bitmap_invalid = TRUE;
				break;

			default:
				break;
		}

	if (bitmap_invalid && im->ImageData)
	{
		DB1(DBPRINTF("VectorGlyph:   Reallocating Image\n");)
		ASSERT_VALID_PTR(im->ImageData)
		FreeBitMap((struct BitMap *)im->ImageData);
		im->ImageData = NULL;
	}

	return DoSuperMethodA(cl, (Object *)im, (Msg)msg);
}



static ULONG VG_OMNew(Class *cl, struct Image *im, struct opSet *msg)
{
	DB2(DBPRINTF("VectorGlyph: OM_NEW\n");)

	if ((im = (struct Image *)DoSuperMethodA(cl, (Object *)im, (Msg) msg)))
	{
		/* Store the glyph type in Image->PlanePick */
		im->PlanePick = GetTagData(SYSIA_Which, 0, msg->ops_AttrList);
		im->ImageData = NULL;
		return (ULONG)im;
	}

	return NULL;
}



static void VG_OMDispose(Class *cl, struct Image *im, Msg msg)
{
	DB2(DBPRINTF("VectorGlyph: OM_DISPOSE\n");)

	/* Restore original depth! */
	/* ((struct BitMap *)im->ImageData)->Depth = 1; */

	FreeBitMap((struct BitMap *)im->ImageData);

	/* Now let our superclass free its istance */
	DoSuperMethodA(cl, (Object *)im, msg);
}



static struct BitMap *CreateGlyph(ULONG which, ULONG width, ULONG height, struct BitMap *friend)
{
	struct RastPort rp;

	DB1(DBPRINTF("VectorGlyph:   CreateGlyph(%ld, %ld, %ld, 0x%lx)\n",
		which, width, height, friend);)


	InitRastPort (&rp);

	/* Create the BitMap */

	rp.BitMap = AllocBitMap(width, height, friend ? friend->Depth : 1,
		BMF_MINPLANES | BMF_CLEAR, friend);

	/* Draw the glyph into the BitMap */

	if (rp.BitMap)
	{
		PLANEPTR		planeptr;
		struct TmpRas	tmpras;
		struct AreaInfo	areainfo;
		WORD			areabuffer[(5 * 10 + 1) / 2];

		/* Allocate the TmpRass needed by AreaDraw */
		if ((planeptr = AllocRaster(width, height)))
		{
			InitTmpRas(&tmpras, planeptr, RASSIZE(width, height));
			InitArea(&areainfo, areabuffer, 10);
			SetAPen(&rp, 1);
			rp.TmpRas = &tmpras;
			rp.AreaInfo = &areainfo;

			if (which < VG_IMGCOUNT)
				(*DrawFuncs[which])(&rp, width, height);

			/* Just to be sure... */
			rp.TmpRas = NULL;
			rp.AreaInfo = NULL;
			FreeRaster(planeptr, width, height);
		}

		/* Set the bitmap to depth 8, and clear all bitplane
		 * pointers except the first one.
		 * This way our image will complement better.
		 * See the BltBitMap() autodoc for the meaning
		 * of NULL bitplane pointers in the BitMap structure.
		 */
#if 0
		{
			int i;

			for (i = 1; i < 8; i++)
				rp.BitMap->Planes[i] = (UBYTE *)0;

			rp.BitMap->Depth = 8;
		}
#endif

		/* NOTE: Failing to allocate the TmpRas will cause the
		 * image to be blank, but no error will be reported.
		 */
	}

	return rp.BitMap;
}



static void DrawPlayImage(struct RastPort *rp, UWORD width, UWORD height)
{
	UWORD	ymin = height / 4,
			ymax = (height * 3) / 4,
			ymid;

	ymin -= (ymax - ymin) & 1;	/* Force odd heigth for better arrow aspect */
	ymid = (ymin + ymax) / 2;

	RectFill(rp, 1, ymin, (width / 4) - 1, ymax);

	AreaMove(rp, width / 3, ymin);
	AreaDraw(rp, width - 2, ymid);
	AreaDraw(rp, width / 3, ymax);

	AreaEnd(rp);
}



static void DrawStopImage (struct RastPort *rp, UWORD width, UWORD height)
{
	RectFill(rp, width / 4, height / 4, (width * 3) / 4, (height * 3) / 4);
}



static void DrawForwardImage(struct RastPort *rp, UWORD width, UWORD height)
{
	UWORD	ymin = height / 4,
			ymax = (height * 3) / 4,
			ymid;

	ymin -= (ymax - ymin) & 1;	/* Force odd heigth for better arrow aspect */
	ymid = (ymin + ymax) / 2;

	AreaMove(rp, 1, ymin);
	AreaDraw(rp, width / 2, ymid);
	AreaDraw(rp, 1, ymax);

	AreaMove(rp, width / 2, ymin);
	AreaDraw(rp, width - 2, ymid);
	AreaDraw(rp, width / 2, ymax);

	AreaEnd(rp);
}



static void DrawRewindImage(struct RastPort *rp, UWORD width, UWORD height)
{
	UWORD	ymin = height / 4,
			ymax = (height * 3) / 4,
			ymid;

	ymin -= (ymax - ymin) & 1;	/* Force odd heigth for better arrow aspect */
	ymid = (ymin + ymax) / 2;

	AreaMove(rp, width - 2, ymin);
	AreaDraw(rp, width / 2, ymid);
	AreaDraw(rp, width - 2, ymax);

	AreaMove(rp, width / 2 - 1, ymin);
	AreaDraw(rp, 1, ymid);
	AreaDraw(rp, width / 2 - 1, ymax);

	AreaEnd(rp);
}



static void DrawPickImage(struct RastPort *rp, UWORD width, UWORD height)
/*
 *      arrowxmin
 *      | tailxmin
 *      | | tailxmax
 *      | | |
 *      | v v
 *      | ###<----tailymin
 *      v ###
 *      #######<--arrowymin
 *       #####
 *        ###
 *         #<-----arrowymax
 *      #######<--arrowymax+1
 */
{
	UWORD	tailymin	= height / 6,
			tailxmin	= (width * 2) / 5,
			tailxmax	= (width * 3) / 5,
			arrowymin	= (height * 2) / 5,
			arrowymax	= (height * 4) / 5,
			arrowxmin	= width / 5,
			arrowxmax	= (width * 4) / 5;

	AreaMove(rp, tailxmin, tailymin);
	AreaDraw(rp, tailxmax, tailymin);
	AreaDraw(rp, tailxmax, arrowymin);
	AreaDraw(rp, arrowxmax, arrowymin);
	AreaDraw(rp, (arrowxmin + arrowxmax) / 2, arrowymax);
	AreaDraw(rp, arrowxmin, arrowymin);
	AreaDraw(rp, tailxmin, arrowymin);
	AreaEnd(rp);

	if (arrowymax < height - 1) arrowymax++;

	Move(rp, arrowxmin, arrowymax);
	Draw(rp, arrowxmax, arrowymax);
}



static void DrawUpArrowImage(struct RastPort *rp, UWORD width, UWORD height)
{
	UWORD ymax = (height * 4) / 5;

	AreaMove (rp, width / 2, height / 5);
	AreaDraw (rp, (width * 4) / 5, ymax);
	AreaDraw (rp, width / 5, ymax);
	AreaEnd (rp);
}



static void DrawDownArrowImage(struct RastPort *rp, UWORD width, UWORD height)
{
	UWORD ymin = height / 5;

	AreaMove(rp, width / 5, ymin);
	AreaDraw(rp, (width * 4) / 5, ymin);
	AreaDraw(rp, width / 2, (height * 4) / 5);
	AreaEnd(rp);
}



static void DrawLeftArrowImage(struct RastPort *rp, UWORD width, UWORD height)
{
	UWORD xmax = (width * 4) / 5;

	AreaMove(rp, xmax, height / 5);
	AreaDraw(rp, xmax, (height * 4) / 5);
	AreaDraw(rp, width / 5, height / 2);
	AreaEnd(rp);
}



static void DrawRightArrowImage(struct RastPort *rp, UWORD width, UWORD height)
{
	UWORD xmin = width / 5;

	AreaMove(rp, xmin, height / 5);
	AreaDraw(rp, (width * 4) / 5, height / 2);
	AreaDraw(rp, xmin, (height * 4) / 5);
	AreaEnd(rp);
}



/* Class support functions */


Class *MakeVectorGlyphIClass(void)
{
	Class *class;

	if(class = MakeClass(VECTORGLYPHCLASS, IMAGECLASS, NULL, 0, 0))
		class->cl_Dispatcher.h_Entry = (ULONG (*)()) VectorGlyphDispatcher;

	return class;
}



BOOL FreeVectorGlyphIClass(Class *class)
{
	ASSERT_VALID_PTR_OR_NULL(class)
	return FreeClass(class);
}



#if (CLASS_FLAVOUR & FLAVOUR_CLASSLIB)
/*
 * Class library support functions
 */

struct ClassLibrary * HOOKCALL _UserLibInit(REG(a6, struct ClassLibrary *mybase))
{
	SysBase = *((struct ExecBase **)4);	/* Initialize SysBase */

	IntuitionBase	= (struct IntuitionBase *) OpenLibrary("intuition.library", 39);
	GfxBase			= (struct GfxBase *) OpenLibrary("graphics.library", 39);
	UtilityBase		= (UTILITYBASETYPE *) OpenLibrary("utility.library", 39);

	if (!(IntuitionBase && GfxBase && UtilityBase))
	{
		_UserLibCleanup(mybase);
		return NULL;
	}

	if (!(mybase->cl_Class = MakeVectorGlyphIClass()))
	{
		_UserLibCleanup(mybase);
		return NULL;
	}

	AddClass(mybase->cl_Class);
	return mybase;
}



struct ClassLibrary * HOOKCALL _UserLibCleanup(REG(a6, struct ClassLibrary *mybase))
{
	if (mybase->cl_Class)
		if (!FreeVectorGlyphIClass(mybase->cl_Class))
			return NULL;

	CloseLibrary((struct Library *)UtilityBase);
	CloseLibrary((struct Library *)GfxBase);
	CloseLibrary((struct Library *)IntuitionBase);

	return mybase;
}



Class * HOOKCALL _GetEngine(REG(a6, struct ClassLibrary *mybase))
{
	return mybase->cl_Class;
}

#endif /* (CLASS_FLAVOUR & FLAVOUR_CLASSLIB) */
